# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_call 1'] = [
    (
        (
            'kp5b8653'
        ,),
        {
            'client_id': 'client_id_0123456789',
            'client_secret': 'client_secret_abcdefghijklmnupqrstuvwxyz',
            'redirect_uri': 'http://localhost:8080/signin',
            'token_url': 'https://github.com/login/oauth/access_token'
        }
    )
]



"""rename table

Revision ID: 12cc2bed191c
Revises: a4105346b478
Create Date: 2021-09-22 15:04:45.975408

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '12cc2bed191c'
down_revision = 'a4105346b478'
branch_labels = None
depends_on = None


def upgrade():
    op.rename_table('attribute_text', 'attribute_unicode_text')

def downgrade():
    op.rename_table('attribute_unicode_text', 'attribute_text')



import pytest

from acondbs.db.sa import sa
from acondbs.models import ProductType, Product


##__________________________________________________________________||
@pytest.fixture
def app(app):
    y = app

    type_map = ProductType(name="map")
    map1 = Product(name="map1", contact="xyz", type_=type_map)  # noqa: F841

    with y.app_context():
        sa.session.add(type_map)
        sa.session.commit()

    yield y


##__________________________________________________________________||
def test_simple(app):
    """A simple test of updating an object"""

    with app.app_context():
        product1 = Product.query.filter_by(name="map1").first()
        assert "xyz" == product1.contact
        product1.contact = "abc"
        sa.session.commit()

    with app.app_context():
        product1 = Product.query.filter_by(name="map1").first()
        assert "abc" == product1.contact


##__________________________________________________________________||



"""make login unique on github_users

Revision ID: 8585129ca524
Revises: bad578ddbf3a
Create Date: 2020-12-26 14:31:39.829514

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '8585129ca524'
down_revision = 'bad578ddbf3a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('github_users', schema=None) as batch_op:
        batch_op.create_unique_constraint(batch_op.f('uq_github_users_login'), ['login'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('github_users', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('uq_github_users_login'), type_='unique')

    # ### end Alembic commands ###



from acondbs import ops

from acondbs.models import ProductFilePath


def test_create(app):
    with app.app_context():
        count = ProductFilePath.query.count()
        model = ops.create_product_file_path(path="/new/path")
        assert model.path == "/new/path"
        ops.commit()
        path_id = model.path_id
        assert path_id

    with app.app_context():
        assert ProductFilePath.query.count() == (count + 1)
        model = ProductFilePath.query.filter_by(path_id=path_id).one()
        assert model.path == "/new/path"


def test_updated(app):
    with app.app_context():
        count = ProductFilePath.query.count()

        model = ProductFilePath.query.first()
        path_id = model.path_id
        assert model.path != "/new/path"

        model = ops.update_product_file_path(path_id=path_id, path="/new/path")
        assert model.path == "/new/path"

        ops.commit()

    with app.app_context():
        assert ProductFilePath.query.count() == count
        model = ProductFilePath.query.filter_by(path_id=path_id).one()
        assert model.path == "/new/path"


def test_delete(app):
    with app.app_context():
        model = ops.create_product_file_path(path="/to/delete")
        ops.commit()
        path_id = model.path_id

    with app.app_context():
        count = ProductFilePath.query.count()
        model = ops.delete_product_file_path(path_id=path_id)
        ops.commit()

    with app.app_context():
        assert ProductFilePath.query.count() == (count - 1)
        model = ProductFilePath.query.filter_by(path_id=path_id).one_or_none()
        assert model is None



from .fragment_log import FRAGMENT_LOG

FRAGMENT_LOG_CONNECTION = '''
fragment fragmentLogConnection on LogConnection {
  totalCount
  edges {
    node {
     ...fragmentLog
    }
  }
}
''' + FRAGMENT_LOG



"""Declare SQLAlchemy ORM models

In this package, ORM (Object-relational mapping) models are declared.
One model is mapped to one table in the DB. Models are declared as
Python classes inheriting the Model class in Flask-SQLAlchemy.

"Declaring Models" in Flask-SQLAlchemy doc:
https://flask-sqlalchemy.palletsprojects.com/en/2.x/models/

"Declare a Mapping" in SQLAlchemy doc:
https://docs.sqlalchemy.org/en/14/orm/tutorial.html#declare-a-mapping

"""

##__________________________________________________________________||
from .product import (  # noqa: F401
    ProductType,
    Product,
    ProductFilePath,
    ProductRelationType,
    ProductRelation,
    AttributeUnicodeText,
    AttributeBoolean,
    AttributeInteger,
    AttributeFloat,
    AttributeDate,
    AttributeDateTime,
    AttributeTime,
    FieldType,  # enum
    saEnumFieldType,  # SQLAlchemy Enum
    Field,
    TypeFieldAssociation,
)

from .github import (  # noqa: F401
    GitHubToken,
    GitHubOrg,
    GitHubUser,
    GitHubOrgMembership,
)

from .account import AccountAdmin  # noqa: F401

from .web import WebConfig  # noqa: F401

from .misc import Log  # noqa: F401


##__________________________________________________________________||
def init_app(app):
    """Initialize the Flask application object

    This function is called by `create_app()` of Flask

    Parameters
    ----------
    app : Flask
        The Flask application object, an instance of `Flask`
    """

    _add_owners_to_db_as_admins(app)
    # remove_git_hub_tokens_with_invalid_decryption_key(app)


##__________________________________________________________________||
def _add_owners_to_db_as_admins(app):
    import sqlalchemy
    from ..db.sa import sa

    # Test if tables are defined. For example, tables are not defined
    # when a migration version is being created.
    with app.app_context():
        try:
            _ = AccountAdmin.query.all()
        except sqlalchemy.exc.OperationalError:
            return

    with app.app_context():

        owners = app.config.get("ACONDBS_OWNERS_GITHUB_LOGINS", "")
        # e.g., "octocat,dojocat"

        owners = {e.strip() for e in owners.split(",")}
        # e.g., {"octocat", "dojocat"}

        # remove empty strings
        owners = {e for e in owners if e}

        # sort so that "admin_id" is deterministic in tests
        owners = sorted(owners)

        for owner in owners:
            if (
                admin := AccountAdmin.query.filter_by(
                    git_hub_login=owner
                ).one_or_none()
            ) is None:
                admin = AccountAdmin(git_hub_login=owner)
                sa.session.add(admin)
        sa.session.commit()


##__________________________________________________________________||
def remove_git_hub_tokens_with_invalid_decryption_key(app):
    from ..db.sa import sa

    with app.app_context():
        token_ids = [
            e[0] for e in sa.session.query(GitHubToken.token_id).all()
        ]
        for token_id in token_ids:
            try:
                token = GitHubToken.query.filter_by(  # noqa: F841
                    token_id=token_id
                ).one()
            except ValueError:
                sql = f"delete from {GitHubToken.__tablename__} where token_id={token_id};"
                sa.engine.execute(sql)


##__________________________________________________________________||



"""add a table web_config

Revision ID: ccd0df2c7da1
Revises: aeec123e508c
Create Date: 2020-10-13 10:36:06.201703

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ccd0df2c7da1'
down_revision = 'aeec123e508c'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('web_config',
    sa.Column('config_id', sa.Integer(), nullable=False),
    sa.Column('head_title', sa.Text(), nullable=True),
    sa.Column('toolbar_title', sa.Text(), nullable=True),
    sa.Column('devtool_loadingstate', sa.Boolean(), nullable=True),
    sa.Column('product_creation_dialog', sa.Boolean(), nullable=True),
    sa.Column('product_update_dialog', sa.Boolean(), nullable=True),
    sa.Column('product_deletion_dialog', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('config_id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('web_config')
    # ### end Alembic commands ###



import csv
from io import StringIO

import datetime

from sqlalchemy import MetaData
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy_utils import EncryptedType

import pytest

from acondbs import create_app
from acondbs.db.ops import convert_data_type_for_insert


##__________________________________________________________________||
sa = SQLAlchemy()


class SampleTable(sa.Model):
    __tablename__ = "sample_table"
    id_ = sa.Column(sa.Integer(), primary_key=True)

    # https://docs.sqlalchemy.org/en/14/core/type_basics.html#generic-types
    text = sa.Column(sa.Text())
    unicode_text = sa.Column(sa.UnicodeText())
    boolean = sa.Column(sa.Boolean())
    integer = sa.Column(sa.Integer())
    float = sa.Column(sa.Float())
    date = sa.Column(sa.Date())
    date_time = sa.Column(sa.DateTime())
    time = sa.Column(sa.Time())
    encrypted = sa.Column(EncryptedType(sa.Text(), "8b5d3d25b3e5"))


##__________________________________________________________________||
@pytest.fixture
def app_with_empty_db():
    database_uri = "sqlite:///:memory:"
    app = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    yield app


@pytest.fixture
def app_with_empty_tables(app_with_empty_db):
    app = app_with_empty_db

    # define tables
    with app.app_context():
        engine = sa.engine
        metadata = MetaData()
        metadata.reflect(bind=engine)
        metadata.drop_all(bind=engine)
        sa.Model.metadata.create_all(engine)
    yield app


##__________________________________________________________________||
params = [
    pytest.param(
        dict(
            text="abcde",
            unicode_text="絵文字😀 😃 😄 😁 😆",
            boolean=False,
            integer=512,
            float=2.34556234,
            date=datetime.date(2021, 10, 7),
            date_time=datetime.datetime(2021, 10, 7, 15, 4, 20),
            time=datetime.time(15, 4, 20),
            encrypted="secret string",
        ),
        id="one",
    ),
    pytest.param(
        dict(
            boolean=True,
        ),
        id="bool-true",
    ),
    pytest.param(
        dict(
            text="",
            unicode_text="",
            boolean=None,
            integer=None,
            float=None,
            date=None,
            date_time=None,
            time=None,
            encrypted=None,
        ),
        id="none",
    ),
]


@pytest.mark.parametrize("data", params)
def test_convert(app_with_empty_tables, data):
    """test convert_data_type_for_insert()"""
    app = app_with_empty_tables

    tbl_name = "sample_table"

    expected = list(data.items())  # e.g., [('text', 'abcde'), ...]
    fields = list(data.keys())  # .e.,g ['text', 'unicode_text', ...]

    # delete all rows from the table
    # The table is not empty! Not clear why!
    with app.app_context():
        SampleTable.query.delete()
        sa.session.commit()

    # enter data
    with app.app_context():
        row = SampleTable(**data)
        sa.session.add(row)
        sa.session.commit()

    # assert the data are committed as they entered
    with app.app_context():
        row = SampleTable.query.one()
        actual = [(f, getattr(row, f)) for f in fields]
        assert actual == expected

    # export to csv as string
    with app.app_context():
        csv_str = _export_tbl_to_csv(tbl_name)

        # empty the table
        SampleTable.query.delete()
        sa.session.commit()

    # import from the csv
    with app.app_context():

        # confirm the table is empty
        assert SampleTable.query.count() == 0

        _import_tbl_from_csv(tbl_name, csv_str)

    # assert
    with app.app_context():
        row = SampleTable.query.one()
        actual = [(f, getattr(row, f)) for f in fields]
        assert actual == expected


def _export_tbl_to_csv(tbl_name):
    result_proxy = sa.session.execute(f"select * from {tbl_name}")
    b = StringIO()
    csv_writer = csv.writer(b, lineterminator="\n")
    csv_writer.writerow(result_proxy.keys())
    csv_writer.writerows(result_proxy)
    ret = b.getvalue()
    b.close()
    return ret


def _import_tbl_from_csv(tbl_name, csv_str):

    engine = sa.engine
    metadata = MetaData()
    metadata.reflect(bind=engine)
    tbl = metadata.tables[tbl_name]

    rows = list(csv.reader(StringIO(csv_str)))
    fields = rows[0]
    rows = rows[1:]

    field_types = [tbl.columns[f].type for f in fields]

    data = [
        {
            f: convert_data_type_for_insert(e, t)
            for f, t, e in zip(fields, field_types, r)
        }
        for r in rows
    ]

    ins = tbl.insert()
    sa.session.execute(ins, data)


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[deep] 1'] = {
    'data': {
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'attributesBoolean': {
                            'edges': [
                            ]
                        },
                        'attributesDate': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'date_produced',
                                            'type_': 'DATE'
                                        },
                                        'value': '2020-02-01'
                                    }
                                }
                            ]
                        },
                        'attributesDateTime': {
                            'edges': [
                            ]
                        },
                        'attributesFloat': {
                            'edges': [
                            ]
                        },
                        'attributesInteger': {
                            'edges': [
                            ]
                        },
                        'attributesTime': {
                            'edges': [
                            ]
                        },
                        'attributesUnicodeText': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'contact',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'value': None
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'produced_by',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'value': None
                                    }
                                }
                            ]
                        },
                        'contact': None,
                        'dateProduced': '2020-02-01',
                        'name': 'map1',
                        'note': None,
                        'paths': {
                            'edges': [
                                {
                                    'node': {
                                        'note': None,
                                        'path': 'site2:/another/way/map1',
                                        'pathId': '1'
                                    }
                                },
                                {
                                    'node': {
                                        'note': 'sample comment',
                                        'path': 'site1:/path/to/map1',
                                        'pathId': '8'
                                    }
                                }
                            ]
                        },
                        'postedBy': None,
                        'postingGitHubUser': None,
                        'producedBy': None,
                        'productId': '1',
                        'relations': {
                            'edges': [
                                {
                                    'node': {
                                        'other': {
                                            'name': 'beam1',
                                            'productId': '4',
                                            'typeId': 2,
                                            'type_': {
                                                'name': 'beam',
                                                'typeId': '2'
                                            }
                                        },
                                        'otherProductId': 4,
                                        'relationId': '1',
                                        'reverse': {
                                            'relationId': '2',
                                            'typeId': 1,
                                            'type_': {
                                                'name': 'parent',
                                                'typeId': '1'
                                            }
                                        },
                                        'reverseRelationId': 2,
                                        'typeId': 2,
                                        'type_': {
                                            'name': 'child',
                                            'typeId': '2'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'other': {
                                            'name': 'beam2',
                                            'productId': '5',
                                            'typeId': 2,
                                            'type_': {
                                                'name': 'beam',
                                                'typeId': '2'
                                            }
                                        },
                                        'otherProductId': 5,
                                        'relationId': '3',
                                        'reverse': {
                                            'relationId': '4',
                                            'typeId': 1,
                                            'type_': {
                                                'name': 'parent',
                                                'typeId': '1'
                                            }
                                        },
                                        'reverseRelationId': 4,
                                        'typeId': 2,
                                        'type_': {
                                            'name': 'child',
                                            'typeId': '2'
                                        }
                                    }
                                }
                            ]
                        },
                        'timePosted': '2021-01-04T14:32:20',
                        'timeUpdated': None,
                        'typeId': 1,
                        'type_': {
                            'name': 'map',
                            'typeId': '1'
                        },
                        'updatedBy': None,
                        'updatingGitHubUser': None
                    }
                },
                {
                    'node': {
                        'attributesBoolean': {
                            'edges': [
                            ]
                        },
                        'attributesDate': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'date_produced',
                                            'type_': 'DATE'
                                        },
                                        'value': '2020-02-10'
                                    }
                                }
                            ]
                        },
                        'attributesDateTime': {
                            'edges': [
                            ]
                        },
                        'attributesFloat': {
                            'edges': [
                            ]
                        },
                        'attributesInteger': {
                            'edges': [
                            ]
                        },
                        'attributesTime': {
                            'edges': [
                            ]
                        },
                        'attributesUnicodeText': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'contact',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'value': None
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'produced_by',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'value': None
                                    }
                                }
                            ]
                        },
                        'contact': None,
                        'dateProduced': '2020-02-10',
                        'name': 'map2',
                        'note': None,
                        'paths': {
                            'edges': [
                                {
                                    'node': {
                                        'note': None,
                                        'path': 'site1:/path/to/map2',
                                        'pathId': '2'
                                    }
                                }
                            ]
                        },
                        'postedBy': None,
                        'postingGitHubUser': None,
                        'producedBy': None,
                        'productId': '2',
                        'relations': {
                            'edges': [
                            ]
                        },
                        'timePosted': '2021-01-04T14:32:20',
                        'timeUpdated': None,
                        'typeId': 1,
                        'type_': {
                            'name': 'map',
                            'typeId': '1'
                        },
                        'updatedBy': None,
                        'updatingGitHubUser': None
                    }
                },
                {
                    'node': {
                        'attributesBoolean': {
                            'edges': [
                            ]
                        },
                        'attributesDate': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'date_produced',
                                            'type_': 'DATE'
                                        },
                                        'value': '2020-03-19'
                                    }
                                }
                            ]
                        },
                        'attributesDateTime': {
                            'edges': [
                            ]
                        },
                        'attributesFloat': {
                            'edges': [
                            ]
                        },
                        'attributesInteger': {
                            'edges': [
                            ]
                        },
                        'attributesTime': {
                            'edges': [
                            ]
                        },
                        'attributesUnicodeText': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'contact',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'value': None
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'produced_by',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'value': None
                                    }
                                }
                            ]
                        },
                        'contact': None,
                        'dateProduced': '2020-03-19',
                        'name': 'map3',
                        'note': None,
                        'paths': {
                            'edges': [
                                {
                                    'node': {
                                        'note': None,
                                        'path': 'site1:/path/to/map3',
                                        'pathId': '3'
                                    }
                                },
                                {
                                    'node': {
                                        'note': None,
                                        'path': 'site2:/another/way/map3',
                                        'pathId': '4'
                                    }
                                }
                            ]
                        },
                        'postedBy': None,
                        'postingGitHubUser': None,
                        'producedBy': None,
                        'productId': '3',
                        'relations': {
                            'edges': [
                            ]
                        },
                        'timePosted': '2021-01-04T14:32:20',
                        'timeUpdated': None,
                        'typeId': 1,
                        'type_': {
                            'name': 'map',
                            'typeId': '1'
                        },
                        'updatedBy': None,
                        'updatingGitHubUser': None
                    }
                },
                {
                    'node': {
                        'attributesBoolean': {
                            'edges': [
                            ]
                        },
                        'attributesDate': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'date_produced',
                                            'type_': 'DATE'
                                        },
                                        'value': '2020-02-05'
                                    }
                                }
                            ]
                        },
                        'attributesDateTime': {
                            'edges': [
                            ]
                        },
                        'attributesFloat': {
                            'edges': [
                            ]
                        },
                        'attributesInteger': {
                            'edges': [
                            ]
                        },
                        'attributesTime': {
                            'edges': [
                            ]
                        },
                        'attributesUnicodeText': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'contact',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'value': None
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'produced_by',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'value': None
                                    }
                                }
                            ]
                        },
                        'contact': None,
                        'dateProduced': '2020-02-05',
                        'name': 'beam1',
                        'note': None,
                        'paths': {
                            'edges': [
                                {
                                    'node': {
                                        'note': None,
                                        'path': 'site1:/path/to/beam1',
                                        'pathId': '5'
                                    }
                                },
                                {
                                    'node': {
                                        'note': None,
                                        'path': 'site2:/another/way/beam1',
                                        'pathId': '6'
                                    }
                                }
                            ]
                        },
                        'postedBy': None,
                        'postingGitHubUser': None,
                        'producedBy': None,
                        'productId': '4',
                        'relations': {
                            'edges': [
                                {
                                    'node': {
                                        'other': {
                                            'name': 'map1',
                                            'productId': '1',
                                            'typeId': 1,
                                            'type_': {
                                                'name': 'map',
                                                'typeId': '1'
                                            }
                                        },
                                        'otherProductId': 1,
                                        'relationId': '2',
                                        'reverse': {
                                            'relationId': '1',
                                            'typeId': 2,
                                            'type_': {
                                                'name': 'child',
                                                'typeId': '2'
                                            }
                                        },
                                        'reverseRelationId': 1,
                                        'typeId': 1,
                                        'type_': {
                                            'name': 'parent',
                                            'typeId': '1'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'other': {
                                            'name': 'beam2',
                                            'productId': '5',
                                            'typeId': 2,
                                            'type_': {
                                                'name': 'beam',
                                                'typeId': '2'
                                            }
                                        },
                                        'otherProductId': 5,
                                        'relationId': '5',
                                        'reverse': {
                                            'relationId': '6',
                                            'typeId': 1,
                                            'type_': {
                                                'name': 'parent',
                                                'typeId': '1'
                                            }
                                        },
                                        'reverseRelationId': 6,
                                        'typeId': 2,
                                        'type_': {
                                            'name': 'child',
                                            'typeId': '2'
                                        }
                                    }
                                }
                            ]
                        },
                        'timePosted': '2021-01-04T14:32:20',
                        'timeUpdated': None,
                        'typeId': 2,
                        'type_': {
                            'name': 'beam',
                            'typeId': '2'
                        },
                        'updatedBy': None,
                        'updatingGitHubUser': None
                    }
                },
                {
                    'node': {
                        'attributesBoolean': {
                            'edges': [
                            ]
                        },
                        'attributesDate': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'date_produced',
                                            'type_': 'DATE'
                                        },
                                        'value': '2020-03-04'
                                    }
                                }
                            ]
                        },
                        'attributesDateTime': {
                            'edges': [
                            ]
                        },
                        'attributesFloat': {
                            'edges': [
                            ]
                        },
                        'attributesInteger': {
                            'edges': [
                            ]
                        },
                        'attributesTime': {
                            'edges': [
                            ]
                        },
                        'attributesUnicodeText': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'contact',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'value': None
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'produced_by',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'value': None
                                    }
                                }
                            ]
                        },
                        'contact': None,
                        'dateProduced': '2020-03-04',
                        'name': 'beam2',
                        'note': None,
                        'paths': {
                            'edges': [
                                {
                                    'node': {
                                        'note': None,
                                        'path': 'site1:/path/to/beam2',
                                        'pathId': '7'
                                    }
                                }
                            ]
                        },
                        'postedBy': None,
                        'postingGitHubUser': None,
                        'producedBy': None,
                        'productId': '5',
                        'relations': {
                            'edges': [
                                {
                                    'node': {
                                        'other': {
                                            'name': 'map1',
                                            'productId': '1',
                                            'typeId': 1,
                                            'type_': {
                                                'name': 'map',
                                                'typeId': '1'
                                            }
                                        },
                                        'otherProductId': 1,
                                        'relationId': '4',
                                        'reverse': {
                                            'relationId': '3',
                                            'typeId': 2,
                                            'type_': {
                                                'name': 'child',
                                                'typeId': '2'
                                            }
                                        },
                                        'reverseRelationId': 3,
                                        'typeId': 1,
                                        'type_': {
                                            'name': 'parent',
                                            'typeId': '1'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'other': {
                                            'name': 'beam1',
                                            'productId': '4',
                                            'typeId': 2,
                                            'type_': {
                                                'name': 'beam',
                                                'typeId': '2'
                                            }
                                        },
                                        'otherProductId': 4,
                                        'relationId': '6',
                                        'reverse': {
                                            'relationId': '5',
                                            'typeId': 2,
                                            'type_': {
                                                'name': 'child',
                                                'typeId': '2'
                                            }
                                        },
                                        'reverseRelationId': 5,
                                        'typeId': 1,
                                        'type_': {
                                            'name': 'parent',
                                            'typeId': '1'
                                        }
                                    }
                                }
                            ]
                        },
                        'timePosted': '2021-01-04T14:32:20',
                        'timeUpdated': None,
                        'typeId': 2,
                        'type_': {
                            'name': 'beam',
                            'typeId': '2'
                        },
                        'updatedBy': None,
                        'updatingGitHubUser': None
                    }
                }
            ]
        }
    }
}

snapshots['test_schema[filters-type_id-first-two] 1'] = {
    'data': {
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1',
                        'productId': '1',
                        'typeId': 1
                    }
                },
                {
                    'node': {
                        'name': 'map2',
                        'productId': '2',
                        'typeId': 1
                    }
                }
            ]
        }
    }
}

snapshots['test_schema[filters-type_id-sort] 1'] = {
    'data': {
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map3',
                        'productId': '3',
                        'typeId': 1
                    }
                },
                {
                    'node': {
                        'name': 'map2',
                        'productId': '2',
                        'typeId': 1
                    }
                },
                {
                    'node': {
                        'name': 'map1',
                        'productId': '1',
                        'typeId': 1
                    }
                }
            ]
        }
    }
}

snapshots['test_schema[filters-type_name-first-two] 1'] = {
    'data': {
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1',
                        'productId': '1',
                        'typeId': 1
                    }
                },
                {
                    'node': {
                        'name': 'map2',
                        'productId': '2',
                        'typeId': 1
                    }
                }
            ]
        }
    }
}

snapshots['test_schema[filters-type_name-sort] 1'] = {
    'data': {
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map3',
                        'productId': '3',
                        'typeId': 1
                    }
                },
                {
                    'node': {
                        'name': 'map2',
                        'productId': '2',
                        'typeId': 1
                    }
                },
                {
                    'node': {
                        'name': 'map1',
                        'productId': '1',
                        'typeId': 1
                    }
                }
            ]
        }
    }
}

snapshots['test_schema[first-two-sort] 1'] = {
    'data': {
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map3',
                        'productId': '3',
                        'typeId': 1
                    }
                },
                {
                    'node': {
                        'name': 'beam2',
                        'productId': '5',
                        'typeId': 2
                    }
                }
            ]
        }
    }
}

snapshots['test_schema[first-two] 1'] = {
    'data': {
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1',
                        'productId': '1',
                        'typeId': 1
                    }
                },
                {
                    'node': {
                        'name': 'map2',
                        'productId': '2',
                        'typeId': 1
                    }
                }
            ]
        }
    }
}

snapshots['test_schema[total-count] 1'] = {
    'data': {
        'allProducts': {
            'totalCount': 3
        }
    }
}



"""add table web_config

Revision ID: 3746a752f344
Revises: 63790f65e724
Create Date: 2021-10-27 16:20:22.446644

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3746a752f344'
down_revision = '63790f65e724'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('web_config',
    sa.Column('id_', sa.Integer(), nullable=False),
    sa.Column('json', sa.Text(), nullable=True),
    sa.PrimaryKeyConstraint('id_', name=op.f('pk_web_config'))
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('web_config')
    # ### end Alembic commands ###



"""rename the table admin_app_token github_admin_app_token

Revision ID: f92f99aa6fdd
Revises: 3f6526e7e5ff
Create Date: 2020-12-21 11:44:44.034531

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f92f99aa6fdd'
down_revision = '3f6526e7e5ff'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.rename_table('admin_app_token', 'github_admin_app_token')
    # op.create_table('github_admin_app_token',
    # sa.Column('token_id', sa.Integer(), nullable=False),
    # sa.Column('token', sqlalchemy_utils.types.encrypted.encrypted_type.EncryptedType(), nullable=True),
    # sa.PrimaryKeyConstraint('token_id')
    # )
    # op.drop_table('admin_app_token')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.rename_table('github_admin_app_token', 'admin_app_token')
    # op.create_table('admin_app_token',
    # sa.Column('token_id', sa.INTEGER(), nullable=False),
    # sa.Column('token', sa.BLOB(), nullable=True),
    # sa.PrimaryKeyConstraint('token_id')
    # )
    # op.drop_table('github_admin_app_token')
    # ### end Alembic commands ###



from pathlib import Path
import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables, import_tables_from_csv_files


from acondbs.models import GitHubToken

from ...constants import SAMPLE_DIR


##__________________________________________________________________||
@pytest.fixture
def app():
    config_path = Path(SAMPLE_DIR, "config.py")
    database_uri = "sqlite:///:memory:"
    app = create_app(
        config_path=config_path, SQLALCHEMY_DATABASE_URI=database_uri
    )
    with app.app_context():
        define_tables()
    yield app


##__________________________________________________________________||
def test_encrypted_field(app):
    csvdir = Path(SAMPLE_DIR, "csv")
    with app.app_context():
        import_tables_from_csv_files(csvdir)
        assert "token123" == GitHubToken.query.one().token


def test_how_to_encrypt_and_decrypt(app):
    unencrypted = "token123"
    encrypted = "aKjGknYDHY39Z2xAaN7+sQ=="

    with app.app_context():
        secret_key = app.config["SECRET_KEY"]  # 'secret_key_test_123'

    from sqlalchemy_utils.types.encrypted.encrypted_type import AesEngine

    engine = AesEngine()
    engine._update_key(secret_key)
    engine._set_padding_mechanism(None)

    assert engine.encrypt(unencrypted) == encrypted
    assert engine.decrypt(encrypted) == unencrypted


##__________________________________________________________________||



import pytest

from ...funcs import assert_query

from ..gql import QUERY_FIELD

HEADERS = {
    "Authorization": "Bearer 0fb8c9e16d6f7c4961c4c49212bf197d79f14080"  # dojocat
}


##__________________________________________________________________||
params = [
    pytest.param(
        {
            #
            "query": QUERY_FIELD,
            "variables": {"fieldId": 1},
        },
        id="by-field-id",
    ),
    pytest.param(
        {
            #
            "query": QUERY_FIELD,
            "variables": {"name": "contact"},
        },
        id="by-name",
    ),
]


##__________________________________________________________________||
@pytest.mark.parametrize("data", params)
@pytest.mark.asyncio
async def test_schema(app, snapshot, data):
    await assert_query(app, snapshot, data, HEADERS)


##__________________________________________________________________||



from ...db.sa import sa


##__________________________________________________________________||
class GitHubOrgMembership(sa.Model):
    __tablename__ = "github_org_memberships"
    entry_id = sa.Column(sa.Integer(), primary_key=True)
    org_id = sa.Column(sa.ForeignKey("github_orgs.org_id"), nullable=False)
    org = sa.relationship(
        "GitHubOrg",
        backref=sa.backref("memberships", cascade="all"),
    )
    member_id = sa.Column(
        sa.ForeignKey("github_users.user_id"), nullable=False
    )
    member = sa.relationship(
        "GitHubUser", backref=sa.backref("memberships", cascade="all")
    )


##__________________________________________________________________||



import graphene

from .. import type_

from .... import ops


##__________________________________________________________________||
class CommonInputFields:
    indef_article = graphene.String(
        description=(
            'The indefinite article placed before the singular noun "'
            'i.e., "a" or "an". '
        )
    )
    singular = graphene.String(
        description=("The singular noun, the relation type name in singular.")
    )
    plural = graphene.String(
        description=("The plural noun, the relation type name in plural.")
    )


class CreateProductRelationTypeInput(
    graphene.InputObjectType, CommonInputFields
):
    """An input to createProductRelationTypes()"""

    name = graphene.String(
        required=True, description=("The name of the relation type")
    )


class UpdateProductRelationTypeInput(
    graphene.InputObjectType, CommonInputFields
):
    """An input to updateProductRelationType()"""


##__________________________________________________________________||
class CreateProductRelationTypes(graphene.Mutation):
    """Create a pair of product relation types"""

    class Arguments:
        type = CreateProductRelationTypeInput(
            required=True, description=("A relation type")
        )
        reverse = CreateProductRelationTypeInput(
            description=("The reverse relation type")
        )
        self_reverse = graphene.Boolean(
            description=("true if the reverse type is the same")
        )

    ok = graphene.Boolean()
    product_relation_type = graphene.Field(lambda: type_.ProductRelationType)

    def mutate(root, info, type, reverse=None, self_reverse=False):
        model = ops.create_product_relation_type(type, reverse, self_reverse)
        ops.commit()
        ok = True
        return CreateProductRelationTypes(product_relation_type=model, ok=ok)


class UpdateProductRelationType(graphene.Mutation):
    """Update a product relation type"""

    class Arguments:
        type_id = graphene.Int(
            required=True,
            description=("The typeId of the product relation type"),
        )
        input = UpdateProductRelationTypeInput(required=True)

    ok = graphene.Boolean()
    product_relation_type = graphene.Field(lambda: type_.ProductRelationType)

    def mutate(root, info, type_id, input):
        model = ops.update_product_relation_type(type_id, **input)
        ops.commit()
        ok = True
        return UpdateProductRelationType(product_relation_type=model, ok=ok)


class DeleteProductRelationTypes(graphene.Mutation):
    """Delete a pair of product relation types"""

    class Arguments:
        type_id = graphene.Int(
            required=True,
            description=(
                "The typeId of a relation type. The reverse relation "
                "type will automatically be deleted."
            ),
        )

    ok = graphene.Boolean()

    def mutate(root, info, type_id):
        ops.delete_product_relation_type(type_id)
        ops.commit()
        ok = True
        return DeleteProductRelationTypes(ok=ok)


##__________________________________________________________________||



import pytest

from sqlalchemy import exc

from acondbs import ops
from acondbs.models import ProductType, TypeFieldAssociation


##__________________________________________________________________||
def test_fixture(app):
    with app.app_context():
        assert ProductType.query.count() == 2


##__________________________________________________________________||
params = [
    pytest.param(None, id="none"),
    pytest.param([], id="empty"),
    pytest.param([2], id="one"),
    pytest.param([1, 3, 4], id="multiple"),
    pytest.param([4, 1, 3], id="unsorted"),
    pytest.param([1, 3, 3, 4, 4, 4], id="duplicate"),
    pytest.param([4, 4, 3, 1, 3], id="unsorted-duplicate"),
]


@pytest.mark.parametrize("field_ids", params)
def test_create(app, field_ids):

    with app.app_context():
        count = ProductType.query.count()
        model = ops.create_product_type(
            name="derived_map",
            field_ids=field_ids,
        )
        assert model.name == "derived_map"
        ops.commit()
        type_id = model.type_id
        assert type_id

    with app.app_context():
        assert ProductType.query.count() == (count + 1)
        model = ProductType.query.filter_by(type_id=type_id).one()
        assert model.name == "derived_map"
        expected_field_ids = sorted(set(field_ids)) if field_ids else []
        actual_field_ids = [f.field.field_id for f in model.fields]
        assert actual_field_ids == expected_field_ids


params = [
    pytest.param([1, 3, 4, 88], id="non-existent"),
]


@pytest.mark.parametrize("field_ids", params)
def test_create_error(app, field_ids):

    with app.app_context():
        count = ProductType.query.count()
        with pytest.raises(exc.NoResultFound):
            model = ops.create_product_type(
                name="derived_map",
                field_ids=field_ids,
            )
            ops.commit()

    with app.app_context():
        assert ProductType.query.count() == count
        model = ProductType.query.filter_by(name="derived_map").one_or_none()
        assert model is None


##__________________________________________________________________||
params = [
    pytest.param(None, [1, 2, 3, 7, 9], id="none"),
    pytest.param([], [], id="empty"),
    pytest.param([2], [2], id="removed"),
    pytest.param([1, 2, 3, 4, 5], [1, 2, 3, 4, 5], id="added"),
    pytest.param([1, 4, 5], [1, 4, 5], id="removed-added"),
    pytest.param(
        [4, 5, 5, 5, 1, 4, 5, 1, 1],
        [1, 4, 5],
        id="removed-added-unsorted-duplicate",
    ),
]


@pytest.mark.parametrize("field_ids, expected_field_ids", params)
def test_update(app, field_ids, expected_field_ids):

    type_id = 2

    with app.app_context():
        count = TypeFieldAssociation.query.count()
        model = ProductType.query.filter_by(type_id=type_id).one()
        expected_len_change = len(model.fields) - len(expected_field_ids)

    with app.app_context():
        model = ops.update_product_type(
            type_id=type_id, name="compass", field_ids=field_ids
        )
        ops.commit()
        assert model.name == "compass"

    with app.app_context():
        model = ProductType.query.filter_by(type_id=type_id).one()
        assert model.name == "compass"
        actual_field_ids = [f.field.field_id for f in model.fields]
        assert actual_field_ids == expected_field_ids

        # assert unused TypeFieldAssociation is deleted
        actual_len_change = count - TypeFieldAssociation.query.count()
        assert expected_len_change == actual_len_change


params = [
    pytest.param([2, 4, 88], id="non-existent"),
]


@pytest.mark.parametrize("field_ids", params)
def test_update_error(app, field_ids):

    type_id = 2

    with app.app_context():
        count = TypeFieldAssociation.query.count()

    with app.app_context():
        with pytest.raises(exc.NoResultFound):
            model = ops.update_product_type(
                type_id=type_id,
                name="compass",
                field_ids=field_ids,
            )
            ops.commit()
            # commit() won't be called because of the exception

    with app.app_context():
        model = ProductType.query.filter_by(type_id=type_id).one()
        assert model.name == "beam"
        actual_field_ids = [f.field.field_id for f in model.fields]
        expected_field_ids = [1, 2, 3, 7, 9]
        assert actual_field_ids == expected_field_ids

        assert TypeFieldAssociation.query.count() == count


##__________________________________________________________________||
def test_delete(app):

    with app.app_context():
        model = ops.create_product_type(
            name="to_be_deleted",
            field_ids=[1, 4, 5],
        )
        ops.commit()
        type_id = model.type_id

    with app.app_context():
        count = ProductType.query.count()
        model = ops.delete_product_type(type_id=type_id)
        ops.commit()

    with app.app_context():
        model = ProductType.query.filter_by(type_id=type_id).one_or_none()
        assert model is None
        assert ProductType.query.count() == (count - 1)


##__________________________________________________________________||



FRAGMENT_PRODUCT_RELATION_TYPE = '''
fragment fragmentProductRelationType on ProductRelationType {
  typeId
  name
  indefArticle
  singular
  plural
  reverse {
    typeId
    name
  }
  relations {
    edges {
      node {
        self_ {
          productId
          name
        }
        other {
          productId
          name
        }
      }
    }
  }
}
'''



"""add fields to github_users

Revision ID: c0d8d6a8f071
Revises: 8585129ca524
Create Date: 2020-12-26 15:07:57.347579

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'c0d8d6a8f071'
down_revision = '8585129ca524'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('github_users', schema=None) as batch_op:
        batch_op.add_column(sa.Column('avatar_url', sa.Text(), nullable=True))
        batch_op.add_column(sa.Column('name', sa.Text(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('github_users', schema=None) as batch_op:
        batch_op.drop_column('name')
        batch_op.drop_column('avatar_url')

    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import GenericRepr, Snapshot


snapshots = Snapshot()

snapshots['test_define_tables_start_with_empty_db 1'] = {
    'account_admins': GenericRepr("Table('account_admins', MetaData(), Column('admin_id', INTEGER(), table=<account_admins>, primary_key=True, nullable=False), Column('git_hub_login', TEXT(), table=<account_admins>, nullable=False), schema=None)"),
    'attribute_boolean': GenericRepr("Table('attribute_boolean', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_boolean>, primary_key=True, nullable=False), Column('value', BOOLEAN(), table=<attribute_boolean>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_boolean>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_boolean>), schema=None)"),
    'attribute_date': GenericRepr("Table('attribute_date', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_date>, primary_key=True, nullable=False), Column('value', DATE(), table=<attribute_date>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_date>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_date>), schema=None)"),
    'attribute_date_time': GenericRepr("Table('attribute_date_time', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_date_time>, primary_key=True, nullable=False), Column('value', DATETIME(), table=<attribute_date_time>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_date_time>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_date_time>), schema=None)"),
    'attribute_float': GenericRepr("Table('attribute_float', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_float>, primary_key=True, nullable=False), Column('value', FLOAT(), table=<attribute_float>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_float>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_float>), schema=None)"),
    'attribute_integer': GenericRepr("Table('attribute_integer', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_integer>, primary_key=True, nullable=False), Column('value', INTEGER(), table=<attribute_integer>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_integer>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_integer>), schema=None)"),
    'attribute_time': GenericRepr("Table('attribute_time', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_time>, primary_key=True, nullable=False), Column('value', TIME(), table=<attribute_time>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_time>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_time>), schema=None)"),
    'attribute_unicode_text': GenericRepr("Table('attribute_unicode_text', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_unicode_text>, primary_key=True, nullable=False), Column('value', TEXT(), table=<attribute_unicode_text>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_unicode_text>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_unicode_text>), schema=None)"),
    'field': GenericRepr("Table('field', MetaData(), Column('field_id', INTEGER(), table=<field>, primary_key=True, nullable=False), Column('name', TEXT(), table=<field>, nullable=False), Column('type_', VARCHAR(length=11), table=<field>, nullable=False), schema=None)"),
    'github_org_memberships': GenericRepr("Table('github_org_memberships', MetaData(), Column('entry_id', INTEGER(), table=<github_org_memberships>, primary_key=True, nullable=False), Column('org_id', INTEGER(), ForeignKey('github_orgs.org_id'), table=<github_org_memberships>, nullable=False), Column('member_id', INTEGER(), ForeignKey('github_users.user_id'), table=<github_org_memberships>, nullable=False), schema=None)"),
    'github_orgs': GenericRepr("Table('github_orgs', MetaData(), Column('org_id', INTEGER(), table=<github_orgs>, primary_key=True, nullable=False), Column('git_hub_id', TEXT(), table=<github_orgs>, nullable=False), Column('login', TEXT(), table=<github_orgs>, nullable=False), Column('avatar_url', TEXT(), table=<github_orgs>), Column('url', TEXT(), table=<github_orgs>), schema=None)"),
    'github_tokens': GenericRepr("Table('github_tokens', MetaData(), Column('token_id', INTEGER(), table=<github_tokens>, primary_key=True, nullable=False), Column('token', BLOB(), table=<github_tokens>), Column('scope', TEXT(), table=<github_tokens>), Column('user_id', INTEGER(), ForeignKey('github_users.user_id'), table=<github_tokens>, nullable=False), Column('time_created', DATETIME(), table=<github_tokens>), schema=None)"),
    'github_users': GenericRepr("Table('github_users', MetaData(), Column('user_id', INTEGER(), table=<github_users>, primary_key=True, nullable=False), Column('git_hub_id', TEXT(), table=<github_users>, nullable=False), Column('login', TEXT(), table=<github_users>, nullable=False), Column('name', TEXT(), table=<github_users>), Column('avatar_url', TEXT(), table=<github_users>), Column('url', TEXT(), table=<github_users>), schema=None)"),
    'log': GenericRepr("Table('log', MetaData(), Column('id_', INTEGER(), table=<log>, primary_key=True, nullable=False), Column('level', TEXT(), table=<log>), Column('message', TEXT(), table=<log>), Column('time', DATETIME(), table=<log>), schema=None)"),
    'product_file_paths': GenericRepr("Table('product_file_paths', MetaData(), Column('path_id', INTEGER(), table=<product_file_paths>, primary_key=True, nullable=False), Column('path', TEXT(), table=<product_file_paths>), Column('note', TEXT(), table=<product_file_paths>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<product_file_paths>), schema=None)"),
    'product_relation_types': GenericRepr("Table('product_relation_types', MetaData(), Column('type_id', INTEGER(), table=<product_relation_types>, primary_key=True, nullable=False), Column('name', TEXT(), table=<product_relation_types>, nullable=False), Column('reverse_type_id', INTEGER(), ForeignKey('product_relation_types.type_id'), table=<product_relation_types>), Column('indef_article', TEXT(), table=<product_relation_types>), Column('singular', TEXT(), table=<product_relation_types>), Column('plural', TEXT(), table=<product_relation_types>), schema=None)"),
    'product_relations': GenericRepr("Table('product_relations', MetaData(), Column('relation_id', INTEGER(), table=<product_relations>, primary_key=True, nullable=False), Column('type_id', INTEGER(), ForeignKey('product_relation_types.type_id'), table=<product_relations>, nullable=False), Column('self_product_id', INTEGER(), ForeignKey('products.product_id'), table=<product_relations>, nullable=False), Column('other_product_id', INTEGER(), ForeignKey('products.product_id'), table=<product_relations>, nullable=False), Column('reverse_relation_id', INTEGER(), ForeignKey('product_relations.relation_id'), table=<product_relations>), schema=None)"),
    'product_types': GenericRepr("Table('product_types', MetaData(), Column('type_id', INTEGER(), table=<product_types>, primary_key=True, nullable=False), Column('name', TEXT(), table=<product_types>, nullable=False), Column('order', INTEGER(), table=<product_types>), Column('indef_article', TEXT(), table=<product_types>), Column('singular', TEXT(), table=<product_types>), Column('plural', TEXT(), table=<product_types>), Column('icon', TEXT(), table=<product_types>), schema=None)"),
    'products': GenericRepr("Table('products', MetaData(), Column('product_id', INTEGER(), table=<products>, primary_key=True, nullable=False), Column('type_id', INTEGER(), ForeignKey('product_types.type_id'), table=<products>, nullable=False), Column('name', TEXT(), table=<products>, nullable=False), Column('contact', TEXT(), table=<products>), Column('date_produced', DATE(), table=<products>), Column('produced_by', TEXT(), table=<products>), Column('time_posted', DATETIME(), table=<products>), Column('posted_by', TEXT(), table=<products>), Column('posting_git_hub_user_id', INTEGER(), ForeignKey('github_users.user_id'), table=<products>), Column('time_updated', DATETIME(), table=<products>), Column('updated_by', TEXT(), table=<products>), Column('updating_git_hub_user_id', INTEGER(), ForeignKey('github_users.user_id'), table=<products>), Column('note', TEXT(), table=<products>), schema=None)"),
    'type_field_association': GenericRepr("Table('type_field_association', MetaData(), Column('iid', INTEGER(), table=<type_field_association>, primary_key=True, nullable=False), Column('type_id', INTEGER(), ForeignKey('product_types.type_id'), table=<type_field_association>), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<type_field_association>, nullable=False), schema=None)"),
    'web_config': GenericRepr("Table('web_config', MetaData(), Column('id_', INTEGER(), table=<web_config>, primary_key=True, nullable=False), Column('json', TEXT(), table=<web_config>), schema=None)")
}

snapshots['test_define_tables_start_with_nonempty_db 1'] = {
    'account_admins': GenericRepr("Table('account_admins', MetaData(), Column('admin_id', INTEGER(), table=<account_admins>, primary_key=True, nullable=False), Column('git_hub_login', TEXT(), table=<account_admins>, nullable=False), schema=None)"),
    'attribute_boolean': GenericRepr("Table('attribute_boolean', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_boolean>, primary_key=True, nullable=False), Column('value', BOOLEAN(), table=<attribute_boolean>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_boolean>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_boolean>), schema=None)"),
    'attribute_date': GenericRepr("Table('attribute_date', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_date>, primary_key=True, nullable=False), Column('value', DATE(), table=<attribute_date>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_date>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_date>), schema=None)"),
    'attribute_date_time': GenericRepr("Table('attribute_date_time', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_date_time>, primary_key=True, nullable=False), Column('value', DATETIME(), table=<attribute_date_time>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_date_time>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_date_time>), schema=None)"),
    'attribute_float': GenericRepr("Table('attribute_float', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_float>, primary_key=True, nullable=False), Column('value', FLOAT(), table=<attribute_float>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_float>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_float>), schema=None)"),
    'attribute_integer': GenericRepr("Table('attribute_integer', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_integer>, primary_key=True, nullable=False), Column('value', INTEGER(), table=<attribute_integer>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_integer>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_integer>), schema=None)"),
    'attribute_time': GenericRepr("Table('attribute_time', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_time>, primary_key=True, nullable=False), Column('value', TIME(), table=<attribute_time>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_time>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_time>), schema=None)"),
    'attribute_unicode_text': GenericRepr("Table('attribute_unicode_text', MetaData(), Column('attribute_id', INTEGER(), table=<attribute_unicode_text>, primary_key=True, nullable=False), Column('value', TEXT(), table=<attribute_unicode_text>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<attribute_unicode_text>, nullable=False), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<attribute_unicode_text>), schema=None)"),
    'field': GenericRepr("Table('field', MetaData(), Column('field_id', INTEGER(), table=<field>, primary_key=True, nullable=False), Column('name', TEXT(), table=<field>, nullable=False), Column('type_', VARCHAR(length=11), table=<field>, nullable=False), schema=None)"),
    'github_org_memberships': GenericRepr("Table('github_org_memberships', MetaData(), Column('entry_id', INTEGER(), table=<github_org_memberships>, primary_key=True, nullable=False), Column('org_id', INTEGER(), ForeignKey('github_orgs.org_id'), table=<github_org_memberships>, nullable=False), Column('member_id', INTEGER(), ForeignKey('github_users.user_id'), table=<github_org_memberships>, nullable=False), schema=None)"),
    'github_orgs': GenericRepr("Table('github_orgs', MetaData(), Column('org_id', INTEGER(), table=<github_orgs>, primary_key=True, nullable=False), Column('git_hub_id', TEXT(), table=<github_orgs>, nullable=False), Column('login', TEXT(), table=<github_orgs>, nullable=False), Column('avatar_url', TEXT(), table=<github_orgs>), Column('url', TEXT(), table=<github_orgs>), schema=None)"),
    'github_tokens': GenericRepr("Table('github_tokens', MetaData(), Column('token_id', INTEGER(), table=<github_tokens>, primary_key=True, nullable=False), Column('token', BLOB(), table=<github_tokens>), Column('scope', TEXT(), table=<github_tokens>), Column('user_id', INTEGER(), ForeignKey('github_users.user_id'), table=<github_tokens>, nullable=False), Column('time_created', DATETIME(), table=<github_tokens>), schema=None)"),
    'github_users': GenericRepr("Table('github_users', MetaData(), Column('user_id', INTEGER(), table=<github_users>, primary_key=True, nullable=False), Column('git_hub_id', TEXT(), table=<github_users>, nullable=False), Column('login', TEXT(), table=<github_users>, nullable=False), Column('name', TEXT(), table=<github_users>), Column('avatar_url', TEXT(), table=<github_users>), Column('url', TEXT(), table=<github_users>), schema=None)"),
    'log': GenericRepr("Table('log', MetaData(), Column('id_', INTEGER(), table=<log>, primary_key=True, nullable=False), Column('level', TEXT(), table=<log>), Column('message', TEXT(), table=<log>), Column('time', DATETIME(), table=<log>), schema=None)"),
    'product_file_paths': GenericRepr("Table('product_file_paths', MetaData(), Column('path_id', INTEGER(), table=<product_file_paths>, primary_key=True, nullable=False), Column('path', TEXT(), table=<product_file_paths>), Column('note', TEXT(), table=<product_file_paths>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<product_file_paths>), schema=None)"),
    'product_relation_types': GenericRepr("Table('product_relation_types', MetaData(), Column('type_id', INTEGER(), table=<product_relation_types>, primary_key=True, nullable=False), Column('name', TEXT(), table=<product_relation_types>, nullable=False), Column('reverse_type_id', INTEGER(), ForeignKey('product_relation_types.type_id'), table=<product_relation_types>), Column('indef_article', TEXT(), table=<product_relation_types>), Column('singular', TEXT(), table=<product_relation_types>), Column('plural', TEXT(), table=<product_relation_types>), schema=None)"),
    'product_relations': GenericRepr("Table('product_relations', MetaData(), Column('relation_id', INTEGER(), table=<product_relations>, primary_key=True, nullable=False), Column('type_id', INTEGER(), ForeignKey('product_relation_types.type_id'), table=<product_relations>, nullable=False), Column('self_product_id', INTEGER(), ForeignKey('products.product_id'), table=<product_relations>, nullable=False), Column('other_product_id', INTEGER(), ForeignKey('products.product_id'), table=<product_relations>, nullable=False), Column('reverse_relation_id', INTEGER(), ForeignKey('product_relations.relation_id'), table=<product_relations>), schema=None)"),
    'product_types': GenericRepr("Table('product_types', MetaData(), Column('type_id', INTEGER(), table=<product_types>, primary_key=True, nullable=False), Column('name', TEXT(), table=<product_types>, nullable=False), Column('order', INTEGER(), table=<product_types>), Column('indef_article', TEXT(), table=<product_types>), Column('singular', TEXT(), table=<product_types>), Column('plural', TEXT(), table=<product_types>), Column('icon', TEXT(), table=<product_types>), schema=None)"),
    'products': GenericRepr("Table('products', MetaData(), Column('product_id', INTEGER(), table=<products>, primary_key=True, nullable=False), Column('type_id', INTEGER(), ForeignKey('product_types.type_id'), table=<products>, nullable=False), Column('name', TEXT(), table=<products>, nullable=False), Column('contact', TEXT(), table=<products>), Column('date_produced', DATE(), table=<products>), Column('produced_by', TEXT(), table=<products>), Column('time_posted', DATETIME(), table=<products>), Column('posted_by', TEXT(), table=<products>), Column('posting_git_hub_user_id', INTEGER(), ForeignKey('github_users.user_id'), table=<products>), Column('time_updated', DATETIME(), table=<products>), Column('updated_by', TEXT(), table=<products>), Column('updating_git_hub_user_id', INTEGER(), ForeignKey('github_users.user_id'), table=<products>), Column('note', TEXT(), table=<products>), schema=None)"),
    'type_field_association': GenericRepr("Table('type_field_association', MetaData(), Column('iid', INTEGER(), table=<type_field_association>, primary_key=True, nullable=False), Column('type_id', INTEGER(), ForeignKey('product_types.type_id'), table=<type_field_association>), Column('field_id', INTEGER(), ForeignKey('field.field_id'), table=<type_field_association>, nullable=False), schema=None)"),
    'web_config': GenericRepr("Table('web_config', MetaData(), Column('id_', INTEGER(), table=<web_config>, primary_key=True, nullable=False), Column('json', TEXT(), table=<web_config>), schema=None)")
}



from .fragment_product_type import FRAGMENT_PRODUCT_TYPE

FRAGMENT_PRODUCT_TYPE_CONNECTION = '''
fragment fragmentProductTypeConnection on ProductTypeConnection {
  totalCount
  edges {
    node {
      ...fragmentProductType
    }
  }
}
''' + FRAGMENT_PRODUCT_TYPE



import textwrap
import json
import traceback

from flask import Blueprint, current_app
from flask_graphql import GraphQLView

from .. import auth, schema, ops
from .graphql_ide import GRAPHIQL_NEWER, GRAPHQL_PLAYGROUND

##__________________________________________________________________||
from flask import request


def format_to_str(data_dict):

    format_item = textwrap.dedent(
        """
        - {key}:
        {value}
        """
    ).lstrip()

    return "\n".join(
        [
            format_item.format(
                key=k,
                value=textwrap.indent(str(v), " " * 4).rstrip(),
            )
            for k, v in data_dict.items()
        ]
    )


class GraphQLView(GraphQLView):
    def dispatch_request(self):
        res = super().dispatch_request()
        # return res

        if isinstance(res, str):
            # e.g, GraphiQL
            return res

        try:
            self._log_response(res)
        except BaseException:
            traceback.print_exc()
        finally:
            return res

    def _log_response(self, res):
        if res.status_code == 200:
            return

        level = "ERROR"

        try:
            msg = self._compose_message(res)
        except BaseException:
            msg = traceback.format_exc()

        # print(msg)
        # print()

        ops.create_log(level=level, message=msg)
        ops.commit()

    def _compose_message(self, res):
        content = {
            "Request": self._format_request_to_str(),
            "Response": self._format_response_to_str(res),
        }
        msg = format_to_str(content)
        return msg

    def _format_request_to_str(self):
        content = {
            "Header": str(request.headers),
            "Data": format_to_str(self.parse_body()),
        }
        msg = format_to_str(content)
        # print(msg)
        return msg

    def _format_response_to_str(self, response):
        content = {
            "Status": str(response.status),
            "Data": textwrap.indent(
                json.dumps(response.get_json(), indent=2),
                " " * 4,
            ),
        }
        msg = format_to_str(content)
        return msg


class GraphQLViewW(GraphQLView):
    """A wrapper of GraphQLView.

    Used to determine arguments to GraphQLView.as_view() for each view
    based on on the configuration and request.

    The usual usage of GraphQLView in add_url_rule() as in the document
    (https://github.com/graphql-python/flask-graphql/tree/v2.0.1#usage) is
    as follows
      bp.add_url_rule('/graphql', view_func=GraphQLView.as_view('graphql', schema=schema, graphiql=True))

    The arguments (schema, graphiql) need to have already been
    determined when the module is imported. In this app, they have not
    because they depend on the configuration and request.

    The __init__() will be called in each view from
    https://github.com/pallets/flask/blob/1.1.2/src/flask/views.py#L88

    The arguments are determined in In __init__() and given to the
    base class GraphQLView.

    """

    def __init__(self, **kwargs):
        kwargs.update(
            {
                "schema": _select_schema(),
                "graphiql": current_app.config.get("ACONDBS_GRAPHIQL", False),
                "graphiql_template": _select_graphiql_template(),
            }
        )
        super().__init__(**kwargs)


##__________________________________________________________________||
bp = Blueprint("graphql", __name__)
bp.add_url_rule("/graphql", view_func=GraphQLViewW.as_view("graphql"))


def init_app(app):
    app.register_blueprint(bp)


##__________________________________________________________________||
def _select_schema():
    if auth.is_admin():
        return schema.schema_admin
    elif auth.is_signed_in():
        return schema.schema_private
    else:
        return schema.schema_public


def _select_graphiql_template():
    template_no = current_app.config.get("ACONDBS_GRAPHIQL_TEMPLATE_NO", None)
    if template_no == 1:
        return GRAPHIQL_NEWER
    elif template_no == 2:
        return GRAPHQL_PLAYGROUND
    else:
        return None


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[filter-off] 1'] = {
    'data': {
        'allGitHubUsers': {
            'edges': [
                {
                    'node': {
                        'avatarUrl': 'avatar.com/user1',
                        'login': 'user1',
                        'memberships': {
                            'edges': [
                                {
                                    'node': {
                                        'org': {
                                            'login': 'org1'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'org': {
                                            'login': 'org2'
                                        }
                                    }
                                }
                            ]
                        },
                        'name': 'User One'
                    }
                },
                {
                    'node': {
                        'avatarUrl': 'avatar.com/user2',
                        'login': 'user2',
                        'memberships': {
                            'edges': [
                                {
                                    'node': {
                                        'org': {
                                            'login': 'org2'
                                        }
                                    }
                                }
                            ]
                        },
                        'name': 'User Two'
                    }
                },
                {
                    'node': {
                        'avatarUrl': 'avatar.com/user3',
                        'login': 'user3',
                        'memberships': {
                            'edges': [
                            ]
                        },
                        'name': 'User Three'
                    }
                }
            ],
            'totalCount': 3
        }
    }
}

snapshots['test_schema[filter-on] 1'] = {
    'data': {
        'allGitHubUsers': {
            'edges': [
                {
                    'node': {
                        'avatarUrl': 'avatar.com/user1',
                        'login': 'user1',
                        'memberships': {
                            'edges': [
                                {
                                    'node': {
                                        'org': {
                                            'login': 'org1'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'org': {
                                            'login': 'org2'
                                        }
                                    }
                                }
                            ]
                        },
                        'name': 'User One'
                    }
                },
                {
                    'node': {
                        'avatarUrl': 'avatar.com/user2',
                        'login': 'user2',
                        'memberships': {
                            'edges': [
                                {
                                    'node': {
                                        'org': {
                                            'login': 'org2'
                                        }
                                    }
                                }
                            ]
                        },
                        'name': 'User Two'
                    }
                }
            ],
            'totalCount': 2
        }
    }
}

snapshots['test_schema[one] 1'] = {
    'data': {
        'allGitHubUsers': {
            'edges': [
                {
                    'node': {
                        'avatarUrl': 'avatar.com/user1',
                        'login': 'user1',
                        'memberships': {
                            'edges': [
                                {
                                    'node': {
                                        'org': {
                                            'login': 'org1'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'org': {
                                            'login': 'org2'
                                        }
                                    }
                                }
                            ]
                        },
                        'name': 'User One'
                    }
                },
                {
                    'node': {
                        'avatarUrl': 'avatar.com/user2',
                        'login': 'user2',
                        'memberships': {
                            'edges': [
                                {
                                    'node': {
                                        'org': {
                                            'login': 'org2'
                                        }
                                    }
                                }
                            ]
                        },
                        'name': 'User Two'
                    }
                },
                {
                    'node': {
                        'avatarUrl': 'avatar.com/user3',
                        'login': 'user3',
                        'memberships': {
                            'edges': [
                            ]
                        },
                        'name': 'User Three'
                    }
                }
            ],
            'totalCount': 3
        }
    }
}



import graphene

from ...github.ops import get_github_oauth_app_info

from ..funcs import get_git_hub_viewer_from_info
from ..filter_ import PFilterableConnectionField
from . import type_

##__________________________________________________________________||
all_git_hub_orgs_field = PFilterableConnectionField(type_.GitHubOrg.connection)
all_git_hub_users_field = PFilterableConnectionField(
    type_.GitHubUser.connection
)
all_git_hub_tokens_field = PFilterableConnectionField(
    type_.GitHubToken.connection
)


##__________________________________________________________________||
def resolve_git_hub_o_auth_app_info(parent, info):
    info = get_github_oauth_app_info()
    return type_.GitHubOAuthAppInfo(
        client_id=info["client_id"],
        authorize_url=info["authorize_url"],
        redirect_uri=info["redirect_uri"],
    )


git_hub_o_auth_app_info_field = graphene.Field(
    type_.GitHubOAuthAppInfo, resolver=resolve_git_hub_o_auth_app_info
)


##__________________________________________________________________||
def resolve_git_hub_viewer(parent, info):
    user = get_git_hub_viewer_from_info(info)
    return user


git_hub_viewer_field = graphene.Field(
    type_.GitHubUser, resolver=resolve_git_hub_viewer
)

##__________________________________________________________________||



from ..fragments import FRAGMENT_PRODUCT_SHALLOW

QUERY_PRODUCT_SHALLOW = (
    """
query ProductShallow($productId: Int, $typeId: Int, $name: String) {
  product(productId: $productId, typeId: $typeId, name: $name) {
    ...fragmentProductShallow
  }
}
"""
    + FRAGMENT_PRODUCT_SHALLOW
)



import pytest
import sqlalchemy

from acondbs.github import ops

##__________________________________________________________________||
def test_success(app, snapshot):
    token = 'token3'
    with app.app_context():
        user = ops.get_user_for_token(token)
    assert 'user2' == user.login

##__________________________________________________________________||
def test_error(app, snapshot):
    token = 'no-such-token'
    with app.app_context():
        with pytest.raises(sqlalchemy.orm.exc.NoResultFound):
            _ = ops.get_user_for_token(token)

##__________________________________________________________________||



"""declare ORM models for tables for products

"""

##__________________________________________________________________||
from .product_type import ProductType  # noqa: F401
from .product import Product  # noqa: F401
from .product_file_path import ProductFilePath  # noqa: F401
from .product_relation_type import ProductRelationType  # noqa: F401
from .product_relation import ProductRelation  # noqa: F401

from .attribute import (  # noqa: F401
    AttributeUnicodeText,
    AttributeBoolean,
    AttributeInteger,
    AttributeFloat,
    AttributeDate,
    AttributeDateTime,
    AttributeTime,
)

from .field import (  # noqa: F401
    FieldType,  # enum
    saEnumFieldType,  # SQLAlchemy Enum
    Field,
)


from .type_field_association import TypeFieldAssociation  # noqa: F401

##__________________________________________________________________||



"""add a field "git_hub_id" to github_users

Revision ID: 9f22f713c4ea
Revises: 2b15f740eed9
Create Date: 2021-01-04 15:12:40.931081

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '9f22f713c4ea'
down_revision = '2b15f740eed9'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('github_users', schema=None) as batch_op:
        batch_op.add_column(sa.Column('git_hub_id', sa.Text(), nullable=False))
        batch_op.create_unique_constraint(batch_op.f('uq_github_users_git_hub_id'), ['git_hub_id'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('github_users', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('uq_github_users_git_hub_id'), type_='unique')
        batch_op.drop_column('git_hub_id')

    # ### end Alembic commands ###



import pytest

from ...funcs import assert_query

from ..gql import QUERY_PRODUCT, QUERY_PRODUCT_SHALLOW

HEADERS = {
    "Authorization": "Bearer 0fb8c9e16d6f7c4961c4c49212bf197d79f14080"  # dojocat
}


##__________________________________________________________________||
params = [
    pytest.param(
        {
            #
            "query": QUERY_PRODUCT,
            "variables": {"productId": 1},
        },
        id="deep",
    ),
    pytest.param(
        {
            #
            "query": QUERY_PRODUCT,
            "variables": {"productId": 9899},
        },
        id="product_id-nonexistent",
    ),
    pytest.param(
        {
            #
            "query": QUERY_PRODUCT_SHALLOW,
            "variables": {"typeId": 1, "name": "map1"},
        },
        id="type_id-name",
    ),
    pytest.param(
        {
            #
            "query": QUERY_PRODUCT_SHALLOW,
            "variables": {"productId": 1, "name": "map1"},
        },
        id="product_id-name",
    ),
    pytest.param(
        {
            #
            "query": QUERY_PRODUCT_SHALLOW,
            "variables": {"productId": 1, "name": "map2"},
        },
        id="product_id-name-nonexistent",
    ),
]


##__________________________________________________________________||
@pytest.mark.parametrize("data", params)
@pytest.mark.asyncio
async def test_schema(app, snapshot, data):
    await assert_query(app, snapshot, data, HEADERS)


##__________________________________________________________________||



"""add table account_admins

Revision ID: e1d390d440d5
Revises: 4e8693037e72
Create Date: 2021-02-14 20:06:13.923464

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e1d390d440d5'
down_revision = '4e8693037e72'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('account_admins',
    sa.Column('admin_id', sa.Integer(), nullable=False),
    sa.Column('github_login', sa.Text(), nullable=False),
    sa.PrimaryKeyConstraint('admin_id', name=op.f('pk_account_admins')),
    sa.UniqueConstraint('github_login', name=op.f('uq_account_admins_github_login'))
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('account_admins')
    # ### end Alembic commands ###



import json

from ...db.sa import sa

from ..funcs import shorten


##__________________________________________________________________||
class WebConfig(sa.Model):
    __tablename__ = "web_config"
    id_ = sa.Column(sa.Integer(), primary_key=True)
    json = sa.Column(sa.Text())

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.id_!r} {self._json_repr!r}>"

    @property
    def _json_repr(self):
        try:
            most_compact = json.dumps(json.loads(self.json), indent=None)
            return shorten(most_compact, width=20)
        except BaseException:
            pass

        try:
            return shorten(repr(self.json), width=20)
        except BaseException:
            pass

        return self.json


##__________________________________________________________________||



from ..fragments import FRAGMENT_PRODUCT_CONNECTION_SHALLOW

QUERY_ALL_PRODUCTS_SHALLOW = (
    """
query AllProductsShallow(
  $filters: ProductFilter
  $sort: [ProductSortEnum]
  $first: Int
) {
  allProducts(filters: $filters, sort: $sort, first: $first) {
    ...fragmentProductConnectionShallow
  }
}
"""
    + FRAGMENT_PRODUCT_CONNECTION_SHALLOW
)



from sqlalchemy.orm import declared_attr
from sqlalchemy.orm import declarative_mixin

from ...db.sa import sa


##__________________________________________________________________||
@declarative_mixin
class AttributeBase:
    attribute_id = sa.Column(sa.Integer(), primary_key=True)

    @declared_attr
    def product_id(self):
        return sa.Column(sa.Integer(), sa.ForeignKey("products.product_id"), nullable=False)  # fmt: skip

    @declared_attr
    def product(self):
        return sa.relationship(
            "Product",
            backref=sa.backref(self.backref_column, cascade="all, delete-orphan"),  # fmt: skip
        )

    @declared_attr
    def field_id(self):
        return sa.Column(sa.Integer(), sa.ForeignKey("field.field_id"), nullable=True)  # fmt: skip
        # TODO: Make nullable False

    @declared_attr
    def field(self):
        return sa.relationship(
            "Field",
            backref=sa.backref(self.backref_column, cascade="all, delete-orphan"),  # fmt: skip
        )

    def __repr__(self):
        field_name = self.field.name if self.field else self.field
        return f"<{self.__class__.__name__} {field_name!r}: {self.value!r}>"


##__________________________________________________________________||
class AttributeUnicodeText(AttributeBase, sa.Model):
    __tablename__ = "attribute_unicode_text"
    backref_column = "attributes_unicode_text"
    value = sa.Column(sa.UnicodeText())


class AttributeBoolean(AttributeBase, sa.Model):
    __tablename__ = "attribute_boolean"
    backref_column = "attributes_boolean"
    value = sa.Column(sa.Boolean())


class AttributeInteger(AttributeBase, sa.Model):
    __tablename__ = "attribute_integer"
    backref_column = "attributes_integer"
    value = sa.Column(sa.Integer())


class AttributeFloat(AttributeBase, sa.Model):
    __tablename__ = "attribute_float"
    backref_column = "attributes_float"
    value = sa.Column(sa.Float())


class AttributeDate(AttributeBase, sa.Model):
    __tablename__ = "attribute_date"
    backref_column = "attributes_date"
    value = sa.Column(sa.Date())


class AttributeDateTime(AttributeBase, sa.Model):
    __tablename__ = "attribute_date_time"
    backref_column = "attributes_date_time"
    value = sa.Column(sa.DateTime())


class AttributeTime(AttributeBase, sa.Model):
    __tablename__ = "attribute_time"
    backref_column = "attributes_time"
    value = sa.Column(sa.Time())


##__________________________________________________________________||



"""GraphQL queries as strings

"""

from .fragments import (  # noqa: F401
    FRAGMENT_LOG,
    FRAGMENT_LOG_CONNECTION,
)

from .queries import (  # noqa: F401
    QUERY_ALL_LOGS,
    QUERY_LOG,
)

from .mutations import (  # noqa: F401
    MUTATION_CREATE_LOG,
    MUTATION_DELETE_LOG,
)



import datetime

from ..models import (
    ProductType,
    Product,
    ProductFilePath,
    ProductRelation,
    ProductRelationType,
    GitHubUser,
)

from ..db.sa import sa


##__________________________________________________________________||
def create_product(
    type_id,
    paths=None,
    relations=None,
    attributes=None,
    posting_git_hub_user_id=None,
    **kwargs
):
    """Create a product"""

    with sa.session.no_autoflush:
        return _create_product(
            type_id,
            paths,
            relations,
            attributes,
            posting_git_hub_user_id,
            **kwargs
        )


def update_product(
    product_id,
    paths=None,
    relations=None,
    attributes=None,
    updating_git_hub_user_id=None,
    **kwargs
):
    """Update a product"""

    with sa.session.no_autoflush:
        return _update_product(
            product_id,
            paths,
            relations,
            attributes,
            updating_git_hub_user_id,
            **kwargs
        )


def delete_product(product_id):
    """Delete a product"""

    model = Product.query.filter_by(product_id=product_id).one()
    sa.session.delete(model)
    return


##__________________________________________________________________||
def uniq_preserving_order(list_):
    # https://stackoverflow.com/a/17016257/7309855
    return list(dict.fromkeys(list_))


def _normalize_paths(paths):
    # e.g., paths = ["  /d/e ", " ", "/a/b/c", "/f/g", "/d/e"]

    ret = [p.strip() for p in paths]  # strip
    # e.g., ["/d/e", "", "/a/b/c", "/f/g", "/d/e"]

    ret = [p for p in ret if p]  # remove empty
    # e.g., ["/d/e", "/a/b/c", "/f/g", "/d/e"]

    ret = uniq_preserving_order(ret)
    # e.g., ["/d/e", "/a/b/c", "/f/g"]

    return ret


##__________________________________________________________________||
def _create_product(
    type_id, paths, relations, attributes, posting_git_hub_user_id, **kwargs
):
    product_type = ProductType.query.filter_by(type_id=type_id).one()

    model = Product(type_=product_type, **kwargs)

    if paths is not None:
        paths = _normalize_paths(paths)
        model.paths = [ProductFilePath(path=p) for p in paths]

    if relations is not None:
        model.relations = [
            ProductRelation(
                type_=ProductRelationType.query.filter_by(
                    type_id=r["type_id"]
                ).one(),
                other=Product.query.filter_by(
                    product_id=r["product_id"]
                ).one(),
            )
            for r in relations
        ]

    if attributes is None:
        attributes = {}

    for association in product_type.fields:
        field = association.field
        value = attributes.get(field.field_id)
        field.type_.attribute_class(
            field=field,
            product=model,
            value=value,
        )

    if posting_git_hub_user_id:
        model.posting_git_hub_user = GitHubUser.query.filter_by(
            user_id=posting_git_hub_user_id
        ).one()

    sa.session.add(model)
    return model


def _update_product(
    product_id,
    paths,
    relations,
    attributes,
    updating_git_hub_user_id,
    **kwargs
):

    model = Product.query.filter_by(product_id=product_id).one()

    if not attributes:
        attributes = {}

    # update paths
    if paths is not None:
        model.paths = _update_paths(model.paths, paths)

    # update relations
    if relations is not None:
        model.relations = _update_relations(model.relations, relations)

    # update scalar fields
    for k, v in kwargs.items():
        setattr(model, k, v)

    # update attributes
    for association in model.type_.fields:
        field = association.field
        try:
            value = attributes[field.field_id]
        except KeyError:
            continue
        attribute_class = field.type_.attribute_class
        query = attribute_class.query.filter_by(product=model, field=field)
        attr = query.one_or_none()
        if attr:
            attr.value = value
        else:
            attribute_class(
                name=field.name, field=field, product=model, value=value
            )

    model.time_updated = datetime.datetime.now()
    if updating_git_hub_user_id:
        GitHubUser.query.filter_by(user_id=updating_git_hub_user_id).one()
    model.updating_git_hub_user_id = updating_git_hub_user_id

    return model


def _update_paths(old, input):
    input = _normalize_paths(input)
    path_dict = {p.path: p for p in old}
    return [
        path_dict[p] if p in path_dict else ProductFilePath(path=p)
        for p in input
    ]


def _update_relations(old, input):

    # sets of tuples (type_id, other_product_id) of ProductRelation
    new_ids = {(i["type_id"], i["product_id"]) for i in input}
    old_ids = {(r.type_id, r.other_product_id) for r in old}

    added_ids = new_ids - old_ids

    model_dict = {(r.type_id, r.other_product_id): r for r in old}
    for id_ in added_ids:
        type_id, product_id = id_
        type_ = ProductRelationType.query.filter_by(type_id=type_id).one()
        other = Product.query.filter_by(product_id=product_id).one()
        model_dict[id_] = ProductRelation(type_=type_, other=other)

    return [model_dict[i] for i in new_ids]


##__________________________________________________________________||



from acondbs.db.sa import sa
from acondbs.models import Log


##__________________________________________________________________||
def test_column(app_empty):
    app = app_empty

    with app.app_context():
        model = Log(level="ERROR", message="an exception occurred")
        sa.session.add(model)
        sa.session.commit()
        id_ = model.id_
        assert id_

    with app.app_context():
        model = Log.query.filter_by(id_=id_).one()
        assert model.level == "ERROR"


def test_repr(app_empty):
    app = app_empty

    model = Log(level="ERROR", message="an exception occurred")
    repr(model)

    with app.app_context():
        sa.session.add(model)
        sa.session.commit()
        id_ = model.id_

    with app.app_context():
        model = Log.query.filter_by(id_=id_).one()
        repr(model)


##__________________________________________________________________||



"""add column reverse_type_id to product_relation_types

Revision ID: 3f4e8bfd8a7f
Revises: d8afc036053a
Create Date: 2020-05-24 20:31:48.005028

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3f4e8bfd8a7f'
down_revision = 'd8afc036053a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('product_relation_types', schema=None) as batch_op:
        batch_op.add_column(sa.Column('reverse_type_id', sa.Integer(), nullable=True))
        batch_op.create_foreign_key('fk_product_relation_types_reverse_type_id_type_id', 'product_relation_types', ['reverse_type_id'], ['type_id'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('product_relation_types', schema=None) as batch_op:
        batch_op.drop_constraint('fk_product_relation_types_reverse_type_id_type_id', type_='foreignkey')
        batch_op.drop_column('reverse_type_id')

    # ### end Alembic commands ###



import pytest

from ...funcs import assert_query

from ..gql import (
    QUERY_ALL_PRODUCT_RELATIONS,
    QUERY_ALL_PRODUCT_RELATIONS_TOTAL_COUNT,
)

HEADERS = {
    "Authorization": "Bearer 0fb8c9e16d6f7c4961c4c49212bf197d79f14080"  # dojocat
}


##__________________________________________________________________||
params = [
    pytest.param(
        {"query": QUERY_ALL_PRODUCT_RELATIONS},
        id="query",
    ),
    pytest.param(
        {"query": QUERY_ALL_PRODUCT_RELATIONS_TOTAL_COUNT},
        id="total-count",
    ),
]


##__________________________________________________________________||
@pytest.mark.parametrize("data", params)
@pytest.mark.asyncio
async def test_schema(app, snapshot, data):
    await assert_query(app, snapshot, data, HEADERS)


##__________________________________________________________________||



"""define filters

Define filters by using graphene-sqlalchemy-filter:
https://github.com/art1415926535/graphene-sqlalchemy-filter

example:
https://github.com/art1415926535/graphene-sqlalchemy-filter/blob/1.10.2/examples/clients_and_records/filters.py

"""
import graphene
from graphene_sqlalchemy_filter import (
    FilterableConnectionField,
    FilterSet,
)

from ..models import (
    Product as ProductModel,
    ProductType as ProductTypeModel,
    GitHubToken as GitHubTokenModel,
    GitHubUser as GitHubUserModel,
)


##__________________________________________________________________||
class ProductFilter(FilterSet):
    type_name = graphene.String()

    class Meta:
        model = ProductModel
        fields = {
            "type_id": [
                "eq",
            ],
        }

    @staticmethod
    def type_name_filter(info, query, value):
        # "Filters that require join":
        # https://github.com/art1415926535/graphene-sqlalchemy-filter/tree/1.10.2#filters-that-require-join
        query = query.join(ProductTypeModel)
        filter_ = ProductTypeModel.name == value
        return query, filter_


class ProductTypeFilter(FilterSet):
    class Meta:
        model = ProductTypeModel
        fields = {}


##__________________________________________________________________||
class GitHubTokenFilter(FilterSet):
    class Meta:
        model = GitHubTokenModel
        fields = {
            "scope": [
                "ilike",
            ],
        }


##__________________________________________________________________||
class GitHubUserFilter(FilterSet):
    org_member = graphene.Boolean()

    class Meta:
        model = GitHubUserModel
        fields = {}

    @staticmethod
    def org_member_filter(info, query, value):
        filter_ = GitHubUserModel.memberships.any() if value else None
        return query, filter_


##__________________________________________________________________||
class PFilterableConnectionField(FilterableConnectionField):
    filters = {
        ProductModel: ProductFilter(),
        ProductTypeModel: ProductTypeFilter(),
        GitHubTokenModel: GitHubTokenFilter(),
        GitHubUserModel: GitHubUserFilter(),
    }


##__________________________________________________________________||



import pytest

from .funcs import assert_mutation

from .misc.gql import QUERY_ALL_LOGS

##__________________________________________________________________||
QUERY = """
{
  noSuchField
}
""".strip()

HEADERS = {
    "Authorization": "Bearer 90b2ee5fed25506df04fd37343bb68d1803dd97f"  # octocat
}

HEADERS_QUERY = {
    "Authorization": "Bearer 90b2ee5fed25506df04fd37343bb68d1803dd97f"  # octocat
}


##__________________________________________________________________||
@pytest.fixture
def app(app_users):
    y = app_users
    yield y


##__________________________________________________________________||
params = [
    pytest.param(
        {
            "query": QUERY,
            "variables": {"var1": 100},
        },
        {"query": QUERY_ALL_LOGS},
        id="one",
    ),
]


@pytest.mark.parametrize("data_mutation, data_query", params)
@pytest.mark.asyncio
async def test_schema(
    app, snapshot, data_mutation, data_query, mock_request_backup_db
):

    success = False
    await assert_mutation(
        app,
        snapshot,
        data_mutation,
        HEADERS,
        data_query,
        HEADERS_QUERY,
        mock_request_backup_db,
        success,
    )


##__________________________________________________________________||



import pytest

from .funcs import assert_query

##__________________________________________________________________||
ALEMBIC_VERSION = """
{
  alembicVersion
}
"""

HEADERS = {"Authorization": "Bearer token123"}  # octocat


##__________________________________________________________________||
params = [
    pytest.param(
        {"query": ALEMBIC_VERSION},
        id="alembicVersion",
    ),
]


@pytest.mark.parametrize("data", params)
@pytest.mark.asyncio
async def test_schema(app, snapshot, data):
    await assert_query(app, snapshot, data, HEADERS)

    # Note: the result is None, i.e., {'data': {'alembicVersion': None }}
    # because the migration is not applied in the tests.


##__________________________________________________________________||



"""add attribute tables

Revision ID: 41e5ff6f2878
Revises: babb78f6f6b3
Create Date: 2021-09-14 14:59:30.551921

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "41e5ff6f2878"
down_revision = "babb78f6f6b3"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "attribute_date",
        sa.Column("attribute_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("product_id", sa.Integer(), nullable=False),
        sa.Column("value", sa.Date(), nullable=True),
        sa.ForeignKeyConstraint(
            ["product_id"],
            ["products.product_id"],
            name=op.f("fk_attribute_date_product_id_products"),
        ),
        sa.PrimaryKeyConstraint(
            "attribute_id", name=op.f("pk_attribute_date")
        ),
    )
    op.create_table(
        "attribute_date_time",
        sa.Column("attribute_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("product_id", sa.Integer(), nullable=False),
        sa.Column("value", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["product_id"],
            ["products.product_id"],
            name=op.f("fk_attribute_date_time_product_id_products"),
        ),
        sa.PrimaryKeyConstraint(
            "attribute_id", name=op.f("pk_attribute_date_time")
        ),
    )
    op.create_table(
        "attribute_text",
        sa.Column("attribute_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("product_id", sa.Integer(), nullable=False),
        sa.Column("value", sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(
            ["product_id"],
            ["products.product_id"],
            name=op.f("fk_attribute_text_product_id_products"),
        ),
        sa.PrimaryKeyConstraint(
            "attribute_id", name=op.f("pk_attribute_text")
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("attribute_text")
    op.drop_table("attribute_date_time")
    op.drop_table("attribute_date")
    # ### end Alembic commands ###



from .fragment_product_shallow import FRAGMENT_PRODUCT_SHALLOW

FRAGMENT_PRODUCT_CONNECTION_SHALLOW = '''
fragment fragmentProductConnectionShallow on ProductConnection {
  edges {
    node {
     ...fragmentProductShallow
    }
  }
}
''' + FRAGMENT_PRODUCT_SHALLOW



import datetime
import sqlalchemy
import pytest

from acondbs.db.sa import sa
from acondbs.models import ProductType, Product


##__________________________________________________________________||
def test_add(app):
    """A simple test of adding an object with a date field"""

    type1 = ProductType(name="type1")

    # time_posted needs to be initialized with a datetime.date
    time_posted = datetime.datetime(2019, 2, 23, 9, 10, 25)
    product1 = Product(name="product1", time_posted=time_posted, type_=type1)

    with app.app_context():
        sa.session.add(product1)
        sa.session.commit()

    with app.app_context():
        product1 = Product.query.filter_by(name="product1").first()

        # The type of the field "time_posted" of Product is "datetime.date"
        assert isinstance(product1.time_posted, datetime.date)

        assert (
            datetime.datetime(2019, 2, 23, 9, 10, 25) == product1.time_posted
        )


##__________________________________________________________________||
def test_add_raise(app):
    """A simple test of adding an object with a wrong type"""

    type1 = ProductType(name="type1")

    # It is not impossible to instnaiate a date field with a wrong
    # type, e.g, str
    product1 = Product(
        name="product1", time_posted="2019-02-13 10:15:21", type_=type1
    )

    with app.app_context():

        # It is also possible to add
        sa.session.add(product1)

        # However, it is not possible to commit
        with pytest.raises(sqlalchemy.exc.StatementError):
            sa.session.commit()


##__________________________________________________________________||



from sqlalchemy import exc

import pytest

from acondbs.db.sa import sa
from acondbs.models import FieldType, Field


##__________________________________________________________________||
def test_repr(app_empty):
    app = app_empty  # noqa: F841

    field1 = Field(name="field1")
    repr(field1)

    field1.type_ = FieldType.UnicodeText
    repr(field1)


##__________________________________________________________________||
def test_commit(app_empty):
    app = app_empty

    with app.app_context():
        field1 = Field(name="field1", type_=FieldType.Date)
        sa.session.add(field1)
        sa.session.commit()

    with app.app_context():
        field1 = Field.query.filter_by(name="field1").one()
        assert field1.field_id
        assert field1.name == "field1"
        assert field1.type_ is FieldType.Date


##__________________________________________________________________||
def test_commit_with_field_id(app_empty):
    app = app_empty

    with app.app_context():
        field1 = Field(field_id=259, name="field1", type_=FieldType.Date)
        sa.session.add(field1)
        sa.session.commit()

    with app.app_context():
        field1 = Field.query.filter_by(field_id=259).one()
        assert field1.name == "field1"
        assert field1.type_ is FieldType.Date


##__________________________________________________________________||
def test_nullable(app_empty):
    app = app_empty  # noqa: F841

    # without type_
    with app.app_context():
        field = Field(name="field")
        sa.session.add(field)
        with pytest.raises(exc.IntegrityError):
            sa.session.commit()

    # without name
    with app.app_context():
        field = Field(type_=FieldType.UnicodeText)
        sa.session.add(field)
        with pytest.raises(exc.IntegrityError):
            sa.session.commit()


##__________________________________________________________________||
def test_enum_by_int(app_empty):
    """Test if an enum can be given by a number

    No. It raises an exception at a commit

    TODO: move this test to tests.models.example
    """
    app = app_empty

    with app.app_context():
        field1 = Field(name="field1", type_=1)
        sa.session.add(field1)
        with pytest.raises(exc.StatementError):
            sa.session.commit()


##__________________________________________________________________||



import datetime

from ...db.sa import sa


##__________________________________________________________________||
class Log(sa.Model):
    __tablename__ = "log"
    id_ = sa.Column(sa.Integer(), primary_key=True)
    level = sa.Column(sa.Text())
    message = sa.Column(sa.Text())
    time = sa.Column(sa.DateTime(), default=lambda: datetime.datetime.now())

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.time}>"


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[all] 1'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'note': None,
                        'path': 'site2:/another/way/map1',
                        'pathId': '1',
                        'product': {
                            'name': 'map1'
                        }
                    }
                },
                {
                    'node': {
                        'note': None,
                        'path': 'site1:/path/to/map2',
                        'pathId': '2',
                        'product': {
                            'name': 'map2'
                        }
                    }
                },
                {
                    'node': {
                        'note': None,
                        'path': 'site1:/path/to/map3',
                        'pathId': '3',
                        'product': {
                            'name': 'map3'
                        }
                    }
                },
                {
                    'node': {
                        'note': None,
                        'path': 'site2:/another/way/map3',
                        'pathId': '4',
                        'product': {
                            'name': 'map3'
                        }
                    }
                },
                {
                    'node': {
                        'note': None,
                        'path': 'site1:/path/to/beam1',
                        'pathId': '5',
                        'product': {
                            'name': 'beam1'
                        }
                    }
                },
                {
                    'node': {
                        'note': None,
                        'path': 'site2:/another/way/beam1',
                        'pathId': '6',
                        'product': {
                            'name': 'beam1'
                        }
                    }
                },
                {
                    'node': {
                        'note': None,
                        'path': 'site1:/path/to/beam2',
                        'pathId': '7',
                        'product': {
                            'name': 'beam2'
                        }
                    }
                },
                {
                    'node': {
                        'note': 'sample comment',
                        'path': 'site1:/path/to/map1',
                        'pathId': '8',
                        'product': {
                            'name': 'map1'
                        }
                    }
                }
            ],
            'totalCount': 8
        }
    }
}



import datetime

from flask import current_app
from sqlalchemy_utils import EncryptedType

from ...db.sa import sa


def encription_key():
    return current_app.config["SECRET_KEY"]


##__________________________________________________________________||
class GitHubToken(sa.Model):
    __tablename__ = "github_tokens"
    token_id = sa.Column(sa.Integer(), primary_key=True)
    token = sa.Column(EncryptedType(sa.Text(), key=encription_key))
    scope = sa.Column(sa.Text())
    user_id = sa.Column(sa.ForeignKey("github_users.user_id"), nullable=False)
    user = sa.relationship(
        "GitHubUser",
        backref=sa.backref("tokens", cascade="all, delete-orphan"),
    )
    time_created = sa.Column(
        sa.DateTime(), default=lambda: datetime.datetime.now()
    )


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[one] 1'] = {
    'data': {
        'allLogs': {
            'edges': [
                {
                    'node': {
                        'id_': '1',
                        'level': 'ERROR',
                        'message': '''- Request:
    - Header:
        Content-Type:: application/json\r
        Authorization: Bearer 90b2ee5fed25506df04fd37343bb68d1803dd97f\r
        Remote-Addr: 127.0.0.1\r
        User-Agent: ASGI-Test-Client\r
        Host: localhost\r
        Content-Type: application/json\r
        Content-Length: 60

    - Data:
        - query:
            {
              noSuchField
            }

        - variables:
            {'var1': 100}

- Response:
    - Status:
        400 BAD REQUEST

    - Data:
            {
              "errors": [
                {
                  "message": "Cannot query field \\"noSuchField\\" on type \\"QueryAdmin\\".",
                  "locations": [
                    {
                      "line": 2,
                      "column": 3
                    }
                  ]
                }
              ]
            }
''',
                        'time': '2021-01-04T14:32:20'
                    }
                }
            ],
            'totalCount': 1
        }
    }
}



from .fragment_field import FRAGMENT_FIELD  # noqa: F401
from .fragment_field_connection import FRAGMENT_FIELD_CONNECTION  # noqa: F401
from .fragment_product import FRAGMENT_PRODUCT  # noqa: F401
from .fragment_product_shallow import FRAGMENT_PRODUCT_SHALLOW  # noqa: F401
from .fragment_product_connection import FRAGMENT_PRODUCT_CONNECTION  # noqa: F401
from .fragment_product_connection_shallow import FRAGMENT_PRODUCT_CONNECTION_SHALLOW  # noqa: F401
from .fragment_product_relation_type import FRAGMENT_PRODUCT_RELATION_TYPE  # noqa: F401
from .fragment_product_relation_type_connection import FRAGMENT_PRODUCT_RELATION_TYPE_CONNECTION  # noqa: F401
from .fragment_product_relation import FRAGMENT_PRODUCT_RELATION  # noqa: F401
from .fragment_product_relation_connection import FRAGMENT_PRODUCT_RELATION_CONNECTION  # noqa: F401
from .fragment_product_type import FRAGMENT_PRODUCT_TYPE  # noqa: F401
from .fragment_product_type_connection import FRAGMENT_PRODUCT_TYPE_CONNECTION  # noqa: F401



"""rename table web_config

Revision ID: 63790f65e724
Revises: 3f804a98ab06
Create Date: 2021-10-26 19:33:50.756208

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "63790f65e724"
down_revision = "3f804a98ab06"
branch_labels = None
depends_on = None


def upgrade():
    op.rename_table("web_config", "web_config_old")


def downgrade():
    op.rename_table("web_config_old", "web_config")



from acondbs.models import FieldType, saEnumFieldType


##__________________________________________________________________||
def test_attribute_class():
    for name, type_ in FieldType.__members__.items():
        # e.g., name = "UnicodeText", type_ = FieldType.UnicodeText
        assert type_.attribute_class.__name__ == f"Attribute{name}"


def test_import_sa_enum():
    assert saEnumFieldType


##__________________________________________________________________||
def test_example():

    # can be instantiated with an int
    type_ = FieldType(1)

    # the same object
    assert type_ is FieldType(1)
    assert type_ is FieldType.UnicodeText

    # access to name and value
    assert type_.name == "UnicodeText"
    assert type_.value == 1


##__________________________________________________________________||



"""Queries to GitHub API
"""

import base64

from .call import call_graphql_api

##__________________________________________________________________||
def org(login, token):
    query = """
      query Organization($login: String!) {
        organization(login: $login) {
          id
          login
          avatarUrl
          url
        }
      }
    """
    variables = { "login": login }
    r = call_graphql_api(query=query, variables=variables, token=token)
    # e.g.,
    #   {
    #       "organization": {
    #           "id": "MDEyOk9yZ2FuaXphdGlvbjc1NjMxODQ0",
    #           "login": "urban-octo-disco",
    #           "avatarUrl": "https://avatars0.githubusercontent.com/u/75631844?v=4",
    #           "url": "https://github.com/urban-octo-disco"
    #       }
    #   }

    r['organization']['id'] = _decode_id(r['organization']['id'])
    # e.g., "012:Organization75631844"

    return r['organization']

##__________________________________________________________________||
def org_members(org_login, token):

    first = 100

    query = """
      query OrganizationMembers($org_login: String!, $first: Int!, $after: String) {
        organization(login: $org_login) {
          login
          membersWithRole(first: $first, after: $after) {
            pageInfo {
              endCursor
              hasNextPage
            }
            edges {
              cursor
              role
              node {
                id
                login
                name
                avatarUrl
                url
              }
            }
          }
        }
      }
    """

    # an example query result:
    #   {'organization': {
    #       'login': 'urban-octo-disco',
    #       'membersWithRole': {
    #           'pageInfo': {
    #               'endCursor': 'Y3Vyc29yOnYyOpHOAO9YQQ==',
    #               'hasNextPage': False
    #           },
    #           'edges': [
    #               {
    #                   'cursor': 'Y3Vyc29yOnYyOpHOABUuMQ==',
    #                   'node': {
    #                       'id': 'MDQ6VXNlcjEzODgwODE=',
    #                       'login': 'TaiSakuma',
    #                       'name': 'Tai Sakuma',
    #                       'avatarUrl': 'https://avatars0.githubusercontent.com/u/1388081?v=4',
    #                       'url': 'https://github.com/TaiSakuma'
    #                   },
    #                   'role': 'MEMBER'
    #               },
    #               {
    #                   'cursor': 'Y3Vyc29yOnYyOpHOAO9YQQ==',
    #                   'node': {
    #                       'id': 'MDQ6VXNlcjE1Njg1Njk3',
    #                       'login': 'tai-sakuma',
    #                       'name': None,
    #                       'avatarUrl': 'https://avatars0.githubusercontent.com/u/15685697?v=4',
    #                       'url': 'https://github.com/tai-sakuma'
    #                   },
    #                   'role': 'ADMIN'
    #               }
    #           ],
    #       }}}

    variables = {"org_login": org_login, "first": first}

    edges = []
    while True:
        r = call_graphql_api(query=query, variables=variables, token=token)
        membersWithRole = r['organization']['membersWithRole']
        edges.extend(membersWithRole['edges'])
        pageInfo = membersWithRole['pageInfo']
        if not pageInfo['hasNextPage']:
            break
        variables["after"] = pageInfo['endCursor']

    for e in edges:
        e['node']['id'] = _decode_id(e['node']['id'])

    return edges

##__________________________________________________________________||
def viewer(token):
    """Return info about the GitHub user for a token

    Parameters
    ----------
    token : str
        An access token, e.g, '4d5dc8b74eccdf65859d6ac64358a3a98300c351'

    Returns
    -------
    dict
        e.g.,
            {
                "id": "04:User583231",
                "login": "octocat",
                "name": "The Octocat",
                "avatarUrl": "https://avatars3.githubusercontent.com/u/583231?u=a59fef2a493e2b67dd13754231daf220c82ba84d&v=4",
                "url": "https://github.com/octocat"
            }
    """

    query = '{ viewer { login id name avatarUrl url } }'
    r = call_graphql_api(query=query, token=token)
    # e.g., https://github.com/octocat
    # {
    #     "viewer": {
    #         "id": "MDQ6VXNlcjU4MzIzMQ==",
    #         "login": "octocat",
    #         "name": "The Octocat",
    #         "avatarUrl": "https://avatars3.githubusercontent.com/u/583231?u=a59fef2a493e2b67dd13754231daf220c82ba84d&v=4",
    #         "url": "https://github.com/octocat"
    #     }
    # }

    viewer = r['viewer']

    viewer['id'] = _decode_id(viewer['id'])
    # e.g., '04:User583231'

    return viewer

##__________________________________________________________________||
def _decode_id(id_):
    """Decode a GitHub user or organization ID returned from GitHub GraphQL API

    Parameters
    ----------
    id_ : str
        An encoded GitHub user or GitHub organization ID returned from
        GitHub GraphQL API, e.g., "MDQ6VXNlcjU4MzIzMQ=="

    Returns
    -------
    str
        The decoded ID, e.g., "04:User583231", "012:Organization75631844"

    """
    return base64.b64decode(id_).decode()

##__________________________________________________________________||




"""add attribute tables

Revision ID: 1583f2de9f39
Revises: 9495d7cb31ec
Create Date: 2021-09-23 20:27:24.421568

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1583f2de9f39'
down_revision = '9495d7cb31ec'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('attribute_boolean',
    sa.Column('attribute_id', sa.Integer(), nullable=False),
    sa.Column('name', sa.Text(), nullable=True),
    sa.Column('value', sa.Boolean(), nullable=True),
    sa.Column('product_id', sa.Integer(), nullable=False),
    sa.Column('field_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['field_id'], ['field.field_id'], name=op.f('fk_attribute_boolean_field_id_field')),
    sa.ForeignKeyConstraint(['product_id'], ['products.product_id'], name=op.f('fk_attribute_boolean_product_id_products')),
    sa.PrimaryKeyConstraint('attribute_id', name=op.f('pk_attribute_boolean'))
    )
    op.create_table('attribute_float',
    sa.Column('attribute_id', sa.Integer(), nullable=False),
    sa.Column('name', sa.Text(), nullable=True),
    sa.Column('value', sa.Float(), nullable=True),
    sa.Column('product_id', sa.Integer(), nullable=False),
    sa.Column('field_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['field_id'], ['field.field_id'], name=op.f('fk_attribute_float_field_id_field')),
    sa.ForeignKeyConstraint(['product_id'], ['products.product_id'], name=op.f('fk_attribute_float_product_id_products')),
    sa.PrimaryKeyConstraint('attribute_id', name=op.f('pk_attribute_float'))
    )
    op.create_table('attribute_integer',
    sa.Column('attribute_id', sa.Integer(), nullable=False),
    sa.Column('name', sa.Text(), nullable=True),
    sa.Column('value', sa.Integer(), nullable=True),
    sa.Column('product_id', sa.Integer(), nullable=False),
    sa.Column('field_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['field_id'], ['field.field_id'], name=op.f('fk_attribute_integer_field_id_field')),
    sa.ForeignKeyConstraint(['product_id'], ['products.product_id'], name=op.f('fk_attribute_integer_product_id_products')),
    sa.PrimaryKeyConstraint('attribute_id', name=op.f('pk_attribute_integer'))
    )
    op.create_table('attribute_time',
    sa.Column('attribute_id', sa.Integer(), nullable=False),
    sa.Column('name', sa.Text(), nullable=True),
    sa.Column('value', sa.Time(), nullable=True),
    sa.Column('product_id', sa.Integer(), nullable=False),
    sa.Column('field_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['field_id'], ['field.field_id'], name=op.f('fk_attribute_time_field_id_field')),
    sa.ForeignKeyConstraint(['product_id'], ['products.product_id'], name=op.f('fk_attribute_time_product_id_products')),
    sa.PrimaryKeyConstraint('attribute_id', name=op.f('pk_attribute_time'))
    )
    with op.batch_alter_table('attribute_date', schema=None) as batch_op:
        batch_op.add_column(sa.Column('field_id', sa.Integer(), nullable=True))
        batch_op.alter_column('name',
               existing_type=sa.TEXT(),
               nullable=True)
        batch_op.create_foreign_key(batch_op.f('fk_attribute_date_field_id_field'), 'field', ['field_id'], ['field_id'])

    with op.batch_alter_table('attribute_date_time', schema=None) as batch_op:
        batch_op.add_column(sa.Column('field_id', sa.Integer(), nullable=True))
        batch_op.alter_column('name',
               existing_type=sa.TEXT(),
               nullable=True)
        batch_op.create_foreign_key(batch_op.f('fk_attribute_date_time_field_id_field'), 'field', ['field_id'], ['field_id'])

    with op.batch_alter_table('attribute_unicode_text', schema=None) as batch_op:
        batch_op.add_column(sa.Column('field_id', sa.Integer(), nullable=True))
        batch_op.alter_column('name',
               existing_type=sa.TEXT(),
               nullable=True)
        batch_op.create_foreign_key(batch_op.f('fk_attribute_unicode_text_field_id_field'), 'field', ['field_id'], ['field_id'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('attribute_unicode_text', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('fk_attribute_unicode_text_field_id_field'), type_='foreignkey')
        batch_op.alter_column('name',
               existing_type=sa.TEXT(),
               nullable=False)
        batch_op.drop_column('field_id')

    with op.batch_alter_table('attribute_date_time', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('fk_attribute_date_time_field_id_field'), type_='foreignkey')
        batch_op.alter_column('name',
               existing_type=sa.TEXT(),
               nullable=False)
        batch_op.drop_column('field_id')

    with op.batch_alter_table('attribute_date', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('fk_attribute_date_field_id_field'), type_='foreignkey')
        batch_op.alter_column('name',
               existing_type=sa.TEXT(),
               nullable=False)
        batch_op.drop_column('field_id')

    op.drop_table('attribute_time')
    op.drop_table('attribute_integer')
    op.drop_table('attribute_float')
    op.drop_table('attribute_boolean')
    # ### end Alembic commands ###



import pytest

from acondbs import auth

##__________________________________________________________________||
params = [
    pytest.param('Bearer {token}', id='no-quote'),
    pytest.param('Bearer "{token}"', id='double-quote'),
    pytest.param("Bearer '{token}'", id='single-quote'),
]

@pytest.mark.parametrize('format_', params)
def test_format(format_, app):
    token = '90b2ee5fed25506df04fd37343bb68d1803dd97f'
    environ_base = {'HTTP_AUTHORIZATION': format_.format(token=token)}
    with app.test_request_context(environ_base=environ_base):
        assert token == auth._get_token_from_http_headers()

##__________________________________________________________________||



import graphene

from alembic.migration import MigrationContext

from ..db.conn import get_db_connection


##__________________________________________________________________||
def resolve_version(parent, info):
    from .. import __version__

    return __version__


version_field = graphene.Field(
    graphene.String,
    description="The version of Acondbs",
    resolver=resolve_version,
)


##__________________________________________________________________||
def resolve_alembic_version(parent, info):
    conn = get_db_connection()
    context = MigrationContext.configure(conn)
    ret = context.get_current_revision()
    # e.g., "35e6ddccd22a"
    return ret


alembic_version_field = graphene.Field(
    graphene.String,
    description="The version of Alembic migration",
    resolver=resolve_alembic_version,
)

##__________________________________________________________________||
